from PyQt5.QtCore import pyqtSlot, Qt, QSize
from PyQt5.QtGui import QCursor, QIcon
from PyQt5.QtWidgets import QVBoxLayout, QSpacerItem, QSizePolicy, QFrame, QMainWindow, QWidget, QDockWidget, \
    QGridLayout, QLabel, QPushButton

import kamzik3
from kamzik3.constants import *
from kamzik3.devices.deviceSession import DeviceSession
from kamzik3.gui.deviceGroupWidget import DeviceGroupWidget
from kamzik3.gui.templates.deviceGroupedListTemplate import Ui_MainWindow
from kamzik3.snippets.snippetsYaml import YamlSerializable


class DeviceGroupedListWidget(Ui_MainWindow, QMainWindow, YamlSerializable):
    previous_selected_tab = 0

    def __init__(self, config, parent=None):
        QMainWindow.__init__(self, parent)
        self.session = kamzik3.session
        self.config = config
        self.detached_tabs = {}
        self.setupUi(self)
        self.button_unlock_controls.hide()
        for group_title, group_config in self.config.get("groups", {}).items():
            self.add_group(group_title, group_config.get("widgets", []), group_config.get("model_image", None))

        self.tab_changed(0)

    def yaml_mapping(self):
        mapping = super().yaml_mapping()
        del mapping["parent"]
        return mapping

    def keyPressEvent(self, key_event):
        if key_event.key() == Qt.Key_Escape:
            self.stop_all_devices()

        QMainWindow.keyPressEvent(self, key_event)

    def add_group(self, title, widgets, image=None):
        grouep_card = DeviceGroupWidget(title, image, parent=self.layout_top_controls)
        self.layout_top_controls.layout().addWidget(grouep_card)
        central_widget = QWidget()
        layout = QVBoxLayout(central_widget)
        layout.setContentsMargins(0, 0, 0, 0)
        layout.setSpacing(0)
        for index, widget in enumerate(widgets):
            layout.addWidget(widget)
            widget.sigShowModel.connect(grouep_card.slot_show_model)
            widget.sigHideModel.connect(grouep_card.slot_hide_model)
            if index != len(widgets) - 1:
                line = QFrame()
                line.setFrameShape(QFrame.HLine)
                line.setFrameShadow(QFrame.Plain)
                line.setStyleSheet("QFrame {color: #f0f0f0; padding-left: 5px; padding-right: 5px}")
                layout.addWidget(line)
            grouep_card.add_device(widget.device)

        spacer = QSpacerItem(0, 0, QSizePolicy.Minimum, QSizePolicy.Expanding)
        layout.addItem(spacer)
        self.tab_device_groups.addTab(central_widget, "  {}  ".format(title))

        def change_group(flag, index=self.tab_device_groups.count() - 1):
            self.group_checked(flag, index)

        grouep_card.button_device_group.clicked.connect(change_group)
        grouep_card.button_group_name.clicked.connect(change_group)

        grouep_card.signal_attach_request.connect(self.slot_attach_tab)
        grouep_card.signal_detach_request.connect(self.slot_detach_tab)

    def get_tab_index(self, tab_id):
        return list(self.config.get("groups", {}).keys()).index(tab_id)

    def group_checked(self, flag, index):
        self.tab_changed(index)

    @pyqtSlot("QString")
    def slot_attach_tab(self, tab_id):
        tab_index = self.get_tab_index(tab_id)
        tab_widget = self.detached_tabs[tab_id]
        self.tab_device_groups.insertTab(tab_index, tab_widget.widget(), "  {}  ".format(tab_id))
        self.tab_device_groups.setCurrentIndex(tab_index)
        self.tab_device_groups.removeTab(tab_index + 1)
        tab_widget.close()
        del self.detached_tabs[tab_id]

    @pyqtSlot("QString")
    def slot_detach_tab(self, tab_id):
        """
        Each Tab can be detached from main window.
        This method detached tab and make it as separate floating window.
        :param tab_id:
        :return:
        """
        tab_index = self.get_tab_index(tab_id)
        tab_widget = self.tab_device_groups.widget(tab_index)

        title_bar_widget = QWidget()
        title_bar_widget.setLayout(QGridLayout())
        title_bar_widget.setSizePolicy(QSizePolicy.Minimum, QSizePolicy.Minimum)
        label = QLabel(" - {}".format(tab_id))
        title_bar_widget.setStyleSheet("QWidget {background-color: #dde9ff}")
        label.setStyleSheet("QLabel {font-weight:bold; color:#003fb4}")

        tab_group_button = self.layout_top_controls.layout().itemAt(tab_index).widget().button_toggle_attach
        attach_button = QPushButton("")
        attach_button.clicked.connect(tab_group_button.clicked)
        attach_button.setMaximumSize(20, 20)
        attach_button.setIconSize(QSize(20, 20))
        attach_button.setFlat(True)
        attach_button.setIcon(QIcon(":/icons/icons/attach.png"))
        attach_button.setCursor(QCursor(Qt.PointingHandCursor))
        attach_button.setToolTip(tab_group_button.toolTip())
        title_bar_widget.layout().addWidget(label, 0, 0)
        title_bar_widget.layout().addWidget(attach_button, 0, 1)

        dock_widget = QDockWidget(parent=self)
        dock_widget.setTitleBarWidget(title_bar_widget)
        dock_widget.setStyleSheet('QDockWidget { font-size: 14px; font-weight: bold; background-color:#fff }')
        dock_widget.setFeatures(QDockWidget.DockWidgetFloatable | QDockWidget.DockWidgetMovable)
        dock_widget.setAllowedAreas(Qt.NoDockWidgetArea)
        dock_widget.setWidget(tab_widget)
        dock_widget.adjustSize()

        dummy_tab = self.get_dummy_tab(tab_index, tab_id)
        self.tab_device_groups.insertTab(tab_index, dummy_tab, "  {}  ".format(tab_id))
        self.tab_device_groups.setCurrentIndex(tab_index)
        self.detached_tabs[tab_id] = dock_widget

        dock_widget.setFloating(True)
        dock_widget.show()

    def get_dummy_tab(self, tab_index, tab_id):
        """
        Create default dummy tab which will server as a placeholder for previously attached tab.
        :param tab_index:
        :param tab_id:
        :return:
        """
        dummy_widget = QWidget()
        layout = QGridLayout(dummy_widget)
        dummy_widget.setLayout(layout)
        button_attach = QPushButton("Click to attach {} group back".format(tab_id))
        button_attach.setSizePolicy(QSizePolicy.Expanding, QSizePolicy.Expanding)
        button_attach.setStyleSheet('font-size: 14px; font-weight: bold;')
        button_attach.setCursor(QCursor(Qt.PointingHandCursor))
        tab_group_button = self.layout_top_controls.layout().itemAt(tab_index).widget().button_toggle_attach
        button_attach.clicked.connect(tab_group_button.clicked)
        layout.addWidget(button_attach)
        return dummy_widget

    @pyqtSlot(int)
    def tab_changed(self, index):
        if index < 0:
            return

        css = 'QPushButton#button_device_group {{border:2px solid {0}; border-bottom:0px; border-left:2px solid {2}; border-right:2px solid {3}; background-color:{1}; margin:0px; padding:5px}}'
        top_layout = self.layout_top_controls.layout()
        top_layout.itemAt(self.previous_selected_tab).widget().button_device_group.setStyleSheet(
            css.format("transparent", "transparent", "transparent", "transparent"))
        self.tab_device_groups.setCurrentIndex(index)
        if index == 0:
            style = ("#d1d1d1", "#dde9ff", "transparent", "#d1d1d1")
        elif index == len(self.tab_device_groups) - 1:
            style = ("#d1d1d1", "#dde9ff", "#d1d1d1", "transparent")
        else:
            style = ("#d1d1d1", "#dde9ff", "#d1d1d1", "#d1d1d1")

        top_layout.itemAt(index).widget().button_device_group.setStyleSheet(css.format(*style))
        self.previous_selected_tab = index

    @pyqtSlot()
    def stop_all_devices(self):
        """
        Call stop method for any device in BUSY status
        :return:
        """
        for device in self.session.devices.values():
            try:
                if isinstance(device, DeviceSession):
                    continue
                if device.is_status(STATUS_BUSY):
                    device.stop()
            except AttributeError:
                # Ignore if device has not implemented stop method
                pass

    @pyqtSlot()
    def lock_widget(self):
        """
        Lock all control widgets.
        It's very usefull if user want to prevent any missclicks
        :return:
        """
        for index in range(len(self.tab_device_groups)):
            self.tab_device_groups.widget(index).setEnabled(False)

        self.button_unlock_controls.show()
        self.button_lock_controls.hide()

    @pyqtSlot()
    def unlock_widget(self):
        """
        Unlock widget and make all control widgets accessible to user
        :return:
        """
        for index in range(len(self.tab_device_groups)):
            self.tab_device_groups.widget(index).setEnabled(True)

        self.button_unlock_controls.hide()
        self.button_lock_controls.show()



import re
import time

import numpy as np
import serial
from bidict import bidict

import kamzik3
from kamzik3.constants import *
from kamzik3.devices.devicePort import DevicePort
from kamzik3.devices.deviceSocket import DeviceSocket
from kamzik3.snippets.snippetsDecorators import expose_method

MEASUREMENT_SCALE = bidict(
    {"U": "Value in 1um", "H": "Value in 0.01mm", "Z": "Value in 0.1mm", "M": "Value in 1mm", "S": "Sensor unit",
     "R": "Raw value"})
MEASUREMENT_FACTORS = {"Value in 1um": 1000, "Value in 0.01mm": 100, "Value in 0.1mm": 10, "Value in 1mm": 1,
                       "Sensor unit": 1, "Raw value": 1}
MEASUREMENT_UNITS = {"Value in 1um": "um", "Value in 0.01mm": "mm", "Value in 0.1mm": "mm", "Value in 1mm": "mm",
                     "Sensor unit": "", "Raw value": ""}


class DeviceBaumerLaserSensorPort(DeviceSocket):
    error_pattern = re.compile('^\{0E[A-Z0-9]+\}$')
    terminator = b"}"
    push_commands_max = 1

    def handle_configuration(self):
        start_at = time.time()

        def _finish_configuration(*_, **__):
            self._config_commands()
            self._config_attributes()
            self.start_polling()
            self.set_status(STATUS_CONFIGURED)
            self.logger.info(u"Device configuration took {} sec.".format(time.time() - start_at))

        self.command("{0V}", callback=_finish_configuration)

    def _init_attributes(self):
        DeviceSocket._init_attributes(self)
        self.create_attribute(ATTR_SOFTWARE_VERSION, readonly=True)
        self.create_attribute(ATTR_HARDWARE_VERSION, readonly=True)
        self.create_attribute(ATTR_MANUFACTURE_DATE, readonly=True)
        self.create_attribute(ATTR_MODES, readonly=True)
        self.create_attribute(ATTR_MEASURING_SCALE, readonly=True, default_type=MEASUREMENT_SCALE.inv.keys(),
                              set_function=self.set_measuring_scale)
        self.create_attribute(ATTR_DISTANCE, default_value=0, default_type=np.float, readonly=True, unit="mm",
                              decimals=4)

    def handle_readout(self, readout_buffer):
        command, output, callback, token = DevicePort.handle_readout(self, readout_buffer)
        is_error_message = self.error_pattern.match(output + "}")
        checksum = output[-2:]
        output = output[2:-2]
        if is_error_message is not None:
            self.handle_command_error(command, readout_buffer)
            command = RESPONSE_ERROR
        elif command == "{0V}":
            output_format = output[2]
            waiting_time_pddm = output[3]
            readout_modes = output[18:]
            self.set_raw_value(ATTR_SOFTWARE_VERSION, output[4:10])
            self.set_raw_value(ATTR_HARDWARE_VERSION, output[10:12])
            self.set_raw_value(ATTR_MANUFACTURE_DATE, output[12:18])
            self.set_raw_value(ATTR_MODES, readout_modes)
            self.set_raw_value(ATTR_MEASURING_SCALE, MEASUREMENT_SCALE[output[1]])
            self.set_attribute([ATTR_DISTANCE, UNIT], MEASUREMENT_UNITS[MEASUREMENT_SCALE[output[1]]])
        elif command == "{0M}":
            values = re.match("^M([MA]+[0-9]+)+$", output)
            for group in values.groups():
                if group[0] == "M":
                    factor = MEASUREMENT_FACTORS[self.get_value(ATTR_MEASURING_SCALE)]
                    distance = int(group[1:]) / factor
                    self.set_raw_value(ATTR_DISTANCE, distance)

        if callback is not None:
            self.handle_readout_callback(callback, command, readout_buffer)
        if token:
            kamzik3.session.publisher.push_message(self.device_id, [command, str(readout_buffer)], token)

    def set_measuring_scale(self, value, callback=None):
        self.logger.info(u"Setting measuring scale to {}".format(value))
        self.set_attribute([ATTR_DISTANCE, UNIT], MEASUREMENT_UNITS[value])
        return self.command("{{0S{}}}".format(MEASUREMENT_SCALE.inv[value]), callback, with_token=True)

    @expose_method()
    def laser_on(self, callback=None):
        self.logger.info(u"Setting laser On")
        return self.command("{0L1}", callback, with_token=True)

    @expose_method()
    def laser_off(self, callback=None):
        self.logger.info(u"Setting laser Off")
        return self.command("{0L0}", callback, with_token=True)

    @expose_method()
    def measure_distance(self, callback=None):
        self.logger.info(u"Measuring distance")
        self.laser_on()
        out = self.command("{0M}", callback, with_token=True)
        self.laser_off()
        return out


class DeviceBaumerLaserSensorSocket(DevicePort, DeviceBaumerLaserSensorPort):
    terminator = b"}"

    def __init__(self, port, baud_rate=38400, parity=serial.PARITY_NONE, stop_bits=serial.STOPBITS_ONE,
                 byte_size=serial.EIGHTBITS, device_id=None, config=None):
        DevicePort.__init__(self, port, baud_rate, parity, stop_bits, byte_size, device_id, config)

    def handshake(self):
        self.logger.info(u"Handshake initiated")
        self.push(b"{0V}")
        time.sleep(0.05)  # Give some time for devices to reply
        if self.serial_port.in_waiting != 0:
            output = self.read_all().strip()
            is_error_message = self.error_pattern.match(output)
            return is_error_message is None
        return False



from PyQt5.QtCore import pyqtSlot, pyqtSignal

from kamzik3.constants import *
from kamzik3.gui.deviceWidget import DeviceWidget
from kamzik3.gui.general.stackableWidget import StackableWidget
from kamzik3.gui.templates.shutterTemplate import Ui_Form
from kamzik3.snippets.snippetsYaml import YamlSerializable


class ShutterWidget(Ui_Form, DeviceWidget, StackableWidget, YamlSerializable):
    sig_opened_status_changed = pyqtSignal("PyQt_PyObject", "PyQt_PyObject")
    sig_auto_mode_changed = pyqtSignal("PyQt_PyObject", "PyQt_PyObject")
    configured = False

    def __init__(self, device, config=None, parent=None):
        self.value_controls = {}
        DeviceWidget.__init__(self, device, config, parent)

    def setupUi(self, Form):
        super().setupUi(Form)
        self.button_close.setVisible(False)
        self.button_open.setVisible(False)
        self.set_status_label(self.device.get_value(ATTR_STATUS))
        self.label_device_name.setText(self.device_id)
        self.setEnabled(False)

    def init_signals(self):
        super().init_signals()
        self.sig_opened_status_changed.connect(self.change_opened_status)
        self.sig_auto_mode_changed.connect(self.set_auto_mode_cb)

    def set_device(self, device):
        DeviceWidget.__init__(self, device)
        self.label_axis_name.setText(device.device_id)
        self.set_status_label(device.get_value(ATTR_STATUS))
        self.value_controls = (
            self.label_auto_mode, self.checkbox_auto_mode, self.button_open, self.button_close)

    def open_shutter(self):
        self.device.open_shutter()

    def close_shutter(self):
        self.device.close_shutter()

    def stop(self):
        self.device.stop()

    @pyqtSlot()
    def set_status_label(self, status):
        self.label_status.setText(status)
        self.label_status.setStyleSheet("QLabel {{background:{}}}".format(self.get_status_color(status)))

    @pyqtSlot()
    def slot_handle_configuration(self):
        DeviceWidget.slot_handle_configuration(self)

        self.device.attach_attribute_callback(ATTR_OPENED, self.sig_opened_status_changed.emit)
        self.device.attach_attribute_callback(ATTR_AUTO_MODE, self.sig_auto_mode_changed.emit)
        self.set_opened(self.device.get_value(ATTR_OPENED))

    @pyqtSlot("bool")
    def slot_set_enabled(self, value):
        self.setEnabled(value)

    @pyqtSlot("PyQt_PyObject")
    def slot_set_status(self, value):
        DeviceWidget.slot_set_status(self, value)

        if value in READY_DEVICE_STATUSES:
            self.setEnabled(True)
        else:
            self.setEnabled(False)
            return

        if value == STATUS_BUSY:
            self.button_stop.setEnabled(True)
            self.button_close.setEnabled(False)
            self.button_open.setEnabled(False)
        else:
            self.button_stop.setEnabled(False)
            self.set_opened(self.device.get_value(ATTR_OPENED))

        self.set_status_label(value)

    @pyqtSlot("PyQt_PyObject", "PyQt_PyObject")
    def change_opened_status(self, key, value):
        if key == VALUE:
            self.set_opened(value)

    def set_opened(self, flag):
        self.label_open_status.setText(u"Opened" if flag else u"Closed")
        self.label_open_status.setStyleSheet("QLabel {{background:{}}}".format(u"red" if flag else u"#68d532"))
        if flag:
            self.button_close.setVisible(True)
            self.button_open.setVisible(False)
            self.button_close.setEnabled(True)
        else:
            self.button_close.setVisible(False)
            self.button_open.setVisible(True)
            self.button_open.setEnabled(True)

    @pyqtSlot("bool")
    def set_auto_mode(self, flag):
        self.device.set_value(ATTR_AUTO_MODE, flag)

    @pyqtSlot("PyQt_PyObject", "PyQt_PyObject")
    def set_auto_mode_cb(self, key, value):
        if key == VALUE:
            self.checkbox_auto_mode.setChecked(value)



from PyQt5.QtCore import pyqtSignal
from PyQt5.QtWidgets import QFrame, QComboBox, QLabel, QHBoxLayout
from pint import UndefinedUnitError

from kamzik3 import units
from kamzik3.constants import UNIT
from kamzik3.snippets.snippetsUnits import get_attribute_unit_range
from kamzik3.snippets.snippetsWidgets import clear_layout


class AttributeUnitWidget(QFrame):
    current_unit = None
    sig_unit_changed = pyqtSignal("QString")
    combox = None

    def __init__(self, attribute, config=None, parent=None):
        self.attribute = attribute
        self.config = config
        if self.config is None:
            self.config = {}
        QFrame.__init__(self, parent=parent)
        self.setLayout(QHBoxLayout())
        self.layout().setContentsMargins(0, 0, 0, 0)
        self.set_unit(attribute[UNIT])

    def set_unit(self, unit):
        if self.combox is not None:
            try:
                self.combox.currentTextChanged.disconnect()
            except TypeError:
                pass
            self.combox = None
        clear_layout(self, self.layout())
        self.layout().addWidget(self.get_unit_widget(unit))
        # self.sig_unit_changed.emit(unit)

    def get_unit_widget(self, unit):
        if unit is not None:
            if unit in units.derived_units or self.config.get("static_unit", False):
                return self._get_static_unit_widget(unit)
            else:
                try:
                    return self._get_combox_unit_widget(unit)
                except (UndefinedUnitError, AttributeError, KeyError):
                    return self._get_static_unit_widget(unit)
        else:
            return QLabel(u"")

    def _get_static_unit_widget(self, unit):
        if unit == u"percent":
            unit = u"%"
        return QLabel(u" {}".format(unit))

    def _get_combox_unit_widget(self, unit):
        self.combox = QComboBox()
        self.combox.addItems(
            get_attribute_unit_range(self.attribute)
        )

        # Signal proxy
        self.combox.currentTextChanged.connect(self.sig_unit_changed.emit)
        self.combox.wheelEvent = lambda *args: None
        index = self.combox.findText("{:~}".format(units.Unit(unit)))
        if index == -1:
            self.combox.setCurrentIndex(0)
        else:
            self.combox.setCurrentIndex(index)
        return self.combox

    def close(self):
        self.sig_unit_changed.disconnect()
        clear_layout(self, self.layout())
        self.attribute = None
        self.config = None



import datetime
import os
import re
import time
from threading import Thread

import numpy as np

from kamzik3 import DeviceError
from kamzik3.constants import *
from kamzik3.devices.device import Device
from kamzik3.snippets.snippetsDecorators import expose_method
from kamzik3.snippets.snippetsUnits import device_units

try:
    import PyTango
    from PyTango import DeviceProxy
except ImportError:
    raise DeviceError("PyTango import error!")

ATTR_TANGO_STATUS = "Tango status"

""" Example of yaml configuration
Lambda: !Device:kamzik3.devices.deviceTango.DeviceTango
    path: tango://mll-lambda.desy.de:10000/mll/lambda/01
    config:
      poll_attributes:
        - [SaveFileName, 300]
        - [FileStartNum, 300]
        - [SaveFilePath, 500]
        - [EnergyThreshold, 1000]
        - [ShutterTime, 1000]
        - [FilePreExt, 1000]
        - [FramesPerFile, 1000]
"""


class DeviceTango(Device, DeviceProxy):

    def __init__(self, path, device_id=None, config=None):
        self.path = path
        self.tango_attributes = []
        self.tango_commands = []
        self.tango_exceptions = (
            PyTango.ConnectionFailed, PyTango.CommunicationFailed, PyTango.DevFailed, PyTango.DeviceUnlocked)
        Device.__init__(self, device_id, config)
        self.connect()

    def connect(self, *args):
        """
        Call only this function to connect devices to port / socket / library / ...
        :param args: connect attributes
        """
        self.connecting = True
        try:
            DeviceProxy.__init__(self, self.path)
        except self.tango_exceptions as e:
            self.handle_connection_error(e)
            return

        self.handle_configuration_event()
        self.connected = True
        self.connecting = False

    def handle_configuration(self):
        start_at = time.time()

        try:
            self.tango_attributes = set(self.get_attribute_list()) - set(self.attributes.keys())
        except self.tango_exceptions as e:
            self.handle_connection_error(e)
            raise e

        self.tango_commands = self.command_list_query()

        for attribute in self.tango_attributes:
            try:
                ta = self.read_attribute(attribute)
            except self.tango_exceptions as e:
                self.handle_command_error(attribute, str(e))
                continue

            def set_function(value, attribute_name=attribute, callback=None):
                try:
                    self.write_attribute(attribute_name, value)
                except self.tango_exceptions + (AttributeError, TypeError) as e:
                    raise DeviceError(e)

            tc = self.get_attribute_config(attribute)

            default_type = "str"
            value = ta.value
            min_value, max_value = None, None
            decimals = 0
            if isinstance(value, np.ndarray):
                continue
            elif tc.data_type == PyTango._tango.CmdArgType.DevLong:
                # Not array
                if tc.max_dim_x == 1 and tc.max_dim_y == 0:
                    default_type = np.float32
                    min_value, max_value = self._get_tango_limits(tc, default_type)
                    value = float(value)
            elif tc.data_type == PyTango._tango.CmdArgType.DevBoolean:
                # Not array
                if tc.max_dim_x == 1 and tc.max_dim_y == 0:
                    default_type = bool
                    value = bool(value)
            elif tc.data_type in (PyTango._tango.CmdArgType.DevDouble, PyTango._tango.CmdArgType.DevFloat):
                # Not array
                if tc.max_dim_x == 1 and tc.max_dim_y == 0:
                    default_type = np.float64
                    min_value, max_value = self._get_tango_limits(tc, default_type)
                    value = float(value)
                    decimals_re = re.search("%[0-9]+\.([0-9]+)f", tc.format)
                    if decimals_re is not None:
                        decimals = int(decimals_re.group(1))
            elif tc.data_type == PyTango._tango.CmdArgType.DevULong:
                # Not array
                if tc.max_dim_x == 1 and tc.max_dim_y == 0:
                    default_type = np.uint32
                    min_value, max_value = self._get_tango_limits(tc, default_type)
                    value = int(value)
            elif tc.data_type == PyTango._tango.CmdArgType.DevShort:
                # Not array
                if tc.max_dim_x == 1 and tc.max_dim_y == 0:
                    default_type = np.int16
                    min_value, max_value = self._get_tango_limits(tc, default_type)
                    value = int(value)
            elif tc.data_type == PyTango._tango.CmdArgType.DevUShort:
                # Not array
                if tc.max_dim_x == 1 and tc.max_dim_y == 0:
                    default_type = np.uint16
                    min_value, max_value = self._get_tango_limits(tc, default_type)
                    value = int(value)
            else:
                value = str(value)

            self.create_attribute(attribute, default_value=value, readonly=not ta.nb_written,
                                  description=tc.description,
                                  default_type=default_type, set_function=set_function, unit=tc.unit,
                                  min_value=min_value, max_value=max_value, decimals=decimals)

        for command in self.tango_commands:
            command_name = command.cmd_name
            # Since this device is already DeviceProxy all methods are callable
            # Just add them to exposed methods and we are done :)
            self.exposed_methods.append((command_name, None))

        self.start_polling()
        self._config_commands()
        self._config_attributes()
        self.set_status(STATUS_CONFIGURED)
        self.logger.info(u"Device configuration took {} sec.".format(time.time() - start_at))

    def start_polling(self):
        super().start_polling()
        self.poll_attribute("State", self.config.get("base polling rate", 100))
        self.poll_command(u"State", self.config.get("base polling rate", 100))

        for poll_attribute in self.config.get("poll_attributes", []):
            tango_attribute, timeout = poll_attribute
            self.poll_attribute(tango_attribute, timeout)
            self.poll_command(tango_attribute, timeout)

    def stop_polling(self):
        super().stop_polling()
        try:
            self.stop_poll_attribute("State")

            for poll_attribute in self.config.get("poll_attributes", []):
                tango_attribute, timeout = poll_attribute
                self.stop_poll_attribute(tango_attribute)
        except self.tango_exceptions:
            pass

    def command(self, command, callback=None, with_token=False, returning=True):
        raise DeviceError("Device does not accept any commands.")

    def command_inout(self, param):
        try:
            return DeviceProxy.command_inout(self, param)
        except self.tango_exceptions as e:
            raise DeviceError(e)

    def send_command(self, commands):
        for command in commands:
            tango_attribute = command[0]
            try:
                polled_attribute = self.attribute_history(tango_attribute, 1)[0]
            except UnicodeDecodeError as e:
                continue
            except self.tango_exceptions as e:
                if e.args[0].reason == "API_CantConnectToDevice":
                    self.handle_connection_error(e)
                    return []
                elif e.args[0].reason == "API_NoDataYet":
                    return []
                else:
                    self.handle_command_error(command, e)
                    continue

            self._set([tango_attribute, VALUE], polled_attribute.value)
            if tango_attribute == "State":
                if polled_attribute.value in (PyTango.DevState.MOVING, PyTango.DevState.RUNNING):
                    self.set_status(STATUS_BUSY)
                elif polled_attribute.value in (PyTango.DevState.ON, PyTango.DevState.STANDBY):
                    self.set_status(STATUS_IDLE)
                elif polled_attribute.value == PyTango.DevState.INIT:
                    self.set_status(STATUS_CONFIGURING)
                elif polled_attribute.value in (PyTango.DevState.DISABLE, PyTango.DevState.UNKNOWN):
                    self.set_status(STATUS_DISCONNECTED)
                else:
                    self.set_status(STATUS_ERROR)
        self.response_timestamp = time.time()
        return []

    def valid_command_format(self, command):
        return True

    @staticmethod
    def _get_tango_limits(tc, type_cast):
        if tc.min_value == "Not specified":
            min_value = None
        else:
            min_value = type_cast(tc.min_value)
        if tc.max_value == "Not specified":
            max_value = None
        else:
            max_value = type_cast(tc.max_value)
        return min_value, max_value


class DeviceTangoLambdaDetector(DeviceTango):

    def handle_configuration(self):
        try:
            super().handle_configuration()
            current_mode = self.attributes["OperatingMode"][VALUE]
            self.create_attribute("OperatingMode", current_mode, default_type=["ContinuousReadWrite", "TwentyFourBit"],
                                  set_function=self.set_mode)
            self.attributes["FrameNumbers"][MIN] = 1
        except self.tango_exceptions:
            # Exception was handled on DeviceTango level
            pass

    @expose_method({"Mode": "Opertaing mode"})
    def set_mode(self, Mode):
        """
        For some reason changing mode is completely blocking action.
        There is no status indication, no documentation.
        We just have to stop polling and waiting for some time.
        To be safe waiting time is set to 10s.
        :param Mode:
        :return:
        """
        if self.get_value("OperatingMode") != Mode:
            def _set():
                self.stop_polling()
                self.set_attribute([ATTR_ENABLED, VALUE], False)
                self.set_status(STATUS_BUSY)
                time.sleep(0.2)
                self.write_attribute_asynch("OperatingMode", Mode)
                self._set(["OperatingMode", VALUE], Mode)
                time.sleep(9)
                self.start_polling()
                self.set_attribute([ATTR_ENABLED, VALUE], True)

            Thread(target=_set).start()

    @expose_method(
        {"Name": "FilePrefix", "Exposure": "ShutterTime", "Frames": "FrameNumbers", "Save": "SaveAllImages"})
    def acquire_frame(self, Name, Exposure, Frames, Save):
        now = datetime.datetime.now()
        self.set_attribute(["TriggerMode", VALUE], 0)
        self.set_attribute(["SaveFilePath", VALUE], self.config.get("shotDir"))
        self.set_attribute(["FilePreExt", VALUE], "_{}_{}".format(self.device_id, now.strftime("%Y-%m-%d_%H:%M:%S")))
        self.set_attribute(["FilePrefix", VALUE], Name)
        self.set_attribute(["FileStartNum", VALUE], 0)
        Exposure = device_units(self, "ShutterTime", Exposure).m
        self.set_attribute(["ShutterTime", VALUE], float(Exposure))
        Frames = device_units(self, "FrameNumbers", Frames).m
        self.set_attribute(["FrameNumbers", VALUE], int(float(Frames)))
        self.set_attribute(["SaveAllImages", VALUE], int(bool(Save)))
        self.StartAcq()

    @expose_method({"Exposure": "ShutterTime"})
    def live_view(self, Exposure):
        Exposure = device_units(self, "ShutterTime", Exposure).m
        self.set_attribute(["ShutterTime", VALUE], float(Exposure))
        self.set_attribute(["SaveAllImages", VALUE], False)
        self.set_attribute(["LiveMode", VALUE], True)
        self.set_attribute(["FrameNumbers", VALUE], 100000)
        self.set_attribute(["TriggerMode", VALUE], 0)
        self.StartAcq()

    @expose_method(
        {"Exposure": "ShutterTime", "Frames": "FrameNumbers"})
    def acquire_scan_frame(self, Exposure, Frames):
        Exposure = float(device_units(self, "ShutterTime", Exposure).m)
        if Exposure != self.get_value("ShutterTime"):
            self.set_attribute(["ShutterTime", VALUE], Exposure)
        Frames = int(float(device_units(self, "FrameNumbers", Frames).m))
        if Frames != self.get_value("FrameNumbers"):
            self.set_attribute(["FrameNumbers", VALUE], Frames)
        time.sleep(1)
        self.StartAcq()

    @expose_method()
    def stop(self):
        self.StopAcq()


class DeviceTangoEigerDetector(DeviceTango):

    def __init__(self, path, file_writer, device_id=None, config=None):
        self.file_writer = file_writer
        DeviceTango.__init__(self, path, device_id, config)

    def handle_configuration(self):
        try:
            super().handle_configuration()
            self.share_group(self.file_writer, None, "FileWriter")
        except self.tango_exceptions:
            # Exception was handled on DeviceTango level
            pass

    def _init_attributes(self):
        DeviceTango._init_attributes(self)
        self.create_attribute("shotDir")
        self.create_attribute("scanDir")
        self.create_attribute("scanFilenamePattern")
        self.create_attribute(ATTR_FRAME_COUNT, default_value=0, readonly=True,
                              description="Internal frame counter",
                              default_type=np.uint64, min_value=0)

    @expose_method(
        {"Name": "FilePrefix", "Exposure": "FrameTime", "Frames": "Nimages"})
    def acquire_frame(self, Name, Exposure, Frames):
        self.logger.info(f"Acquisition of {Frames} @ {Exposure} has started.")
        now = datetime.datetime.now()
        Frames = device_units(self, "Nimages", Frames).m
        self.set_attribute(["TriggerMode", VALUE], 3)
        path = os.path.join(self.get_value("shotDir"), f"{Name}_{self.device_id}_{now.strftime('%Y-%m-%d_%H:%M:%S')}")
        self.set_attribute(["FileWriter", "Mode", VALUE], 1)
        self.set_attribute(["FileWriter", "NamePattern", VALUE], path)
        self.set_attribute(["FileWriter", "NimagesPerFile", VALUE], int(Frames))
        Exposure = device_units(self, "FrameTime", Exposure).m
        # According to Jan 10ms is lowest frame time
        frame_time = 10e-3 if Exposure < 10e-3 else Exposure
        self.set_attribute(["FrameTime", VALUE], frame_time)
        self.set_attribute(["CountTime", VALUE], float(Exposure))
        self.set_attribute(["Ntrigger", VALUE], int(1))
        self.set_attribute(["Nimages", VALUE], int(Frames))
        self.Arm()
        self.Trigger()
        self.set_value(ATTR_FRAME_COUNT, self.get_value(ATTR_FRAME_COUNT) + int(Frames))

    @expose_method(
        {"Exposure": "FrameTime", "Frames": "Nimages"})
    def acquire_scan_frame(self, Exposure, Frames):
        Exposure = float(device_units(self, "FrameTime", Exposure).m)
        if Exposure != self.get_value("CountTime"):
            frame_time = 10e-3 if Exposure < 10e-3 else Exposure
            self.set_attribute(["FrameTime", VALUE], frame_time)
            self.set_attribute(["CountTime", VALUE], Exposure)
        Frames = int(float(device_units(self, "Nimages", Frames).m))
        if Frames != self.get_value("Nimages"):
            self.set_attribute(["Nimages", VALUE], Frames)
            self.set_attribute(["FileWriter", "NimagesPerFile", VALUE], int(Frames))
        self.set_attribute(["FileWriter", "Mode", VALUE], 1)
        self.set_attribute(["FileWriter", "NamePattern", VALUE],
                           self.get_value("scanFilenamePattern").replace("{$ATTR_FRAME_COUNT}",
                                                                         f"{self.get_value(ATTR_FRAME_COUNT):08d}"))
        self.Arm()
        self.Trigger()
        self.set_value(ATTR_FRAME_COUNT, self.get_value(ATTR_FRAME_COUNT) + int(Frames))

    @expose_method(
        {"Exposure": "FrameTime"})
    def live_view(self, Exposure):
        self.logger.info(f"Live view has started.")
        self.set_attribute(["TriggerMode", VALUE], 3)
        self.set_attribute(["FileWriter", "Mode", VALUE], 0)
        self.set_attribute(["FileWriter", "NimagesPerFile", VALUE], 1000000)
        Exposure = device_units(self, "FrameTime", Exposure).m
        # According to Jan 10ms is lowest frame time
        frame_time = 10e-3 if Exposure < 10e-3 else Exposure
        self.set_attribute(["FrameTime", VALUE], frame_time)
        self.set_attribute(["CountTime", VALUE], float(Exposure))
        self.set_attribute(["Ntrigger", VALUE], int(1))
        self.set_attribute(["Nimages", VALUE], 1000000)
        self.Arm()
        self.Trigger()
        self.set_value(ATTR_FRAME_COUNT, self.get_value(ATTR_FRAME_COUNT) + 1000000)

    @expose_method()
    def stop(self):
        if self.get_value(ATTR_STATUS) == STATUS_BUSY:
            self.logger.info(f"Acquisition aborted")
            self.Abort()



from math import inf

from PyQt5.QtCore import pyqtSlot
from PyQt5.QtWidgets import QSizePolicy, QAbstractSpinBox
from pint import UndefinedUnitError, DimensionalityError

from kamzik3 import units
from kamzik3.constants import *
from kamzik3.gui.attributeWidgets.attributeWidget import AttributeWidget
from kamzik3.snippets.snippetsUnits import get_decimals_diff
from kamzik3.snippets.snippetsWidgets import CustomDoubleSpinBox


class AttributeFloatWidget(AttributeWidget):
    attribute_type_cast = float

    def _set_input_widget(self):
        self.input_widget = CustomDoubleSpinBox()
        self.input_widget.setDecimals(self.config.get("max-decimals", self.attribute[DECIMALS]))
        self.input_widget.valueChanged.connect(self.on_numerical_value_set)
        self.input_widget.setSizePolicy(QSizePolicy.Expanding, QSizePolicy.Fixed)

        if self.attribute[READONLY]:
            self.input_widget.setReadOnly(True)
            self.input_widget.setButtonSymbols(QAbstractSpinBox.NoButtons)
        else:
            self.background_color = "#e6ffd9"
            self.input_widget.setStyleSheet(
                ".CustomDoubleSpinBox, .CustomSpinBox {{background:{}; color:{}}}".format(self.background_color,
                                                                                          self.font_color))
        value_min = value_max = "Not set"
        if self.attribute[MIN] is not None:
            value_min = self.attribute.minimum()
            self.input_widget.setMinimum(value_min.m)
        if self.attribute[MAX] is not None:
            value_max = self.attribute.maximum()
            self.input_widget.setMaximum(value_max.m)

        self.set_tooltip("Min: {:~}\nMax: {:~}".format(value_min, value_max))

    def set_value(self, value):
        if value is None or value == u"None":
            self.input_widget.lineEdit().setText("None")
            return

        assert isinstance(value, units.Quantity)

        try:
            set_value = value.to(self.unit)
        except AttributeError:
            set_value = units.Quantity(value)
        self.on_numerical_value_set(set_value.m)
        self.input_widget.setValue(set_value.m)

    def get_attribute_value(self):
        try:
            return units.Quantity(self.input_widget.value(), self.unit).to(self.attribute[UNIT])
        except (UndefinedUnitError, AttributeError):
            return units.Quantity(self.input_widget.value())

    def get_widget_value(self):
        try:
            return units.Quantity(self.input_widget.value(), self.unit)
        except (UndefinedUnitError, AttributeError):
            return units.Quantity(self.input_widget.value())

    def get_minimum(self):
        try:
            return units.Quantity(self.input_widget.minimum(), self.unit)
        except (UndefinedUnitError, AttributeError):
            return units.Quantity(self.input_widget.minimum())

    def get_maximum(self):
        try:
            return units.Quantity(self.input_widget.maximum(), self.unit)
        except (UndefinedUnitError, AttributeError):
            return units.Quantity(self.input_widget.maximum())

    def set_minimum(self, value):
        assert isinstance(value, units.Quantity)
        self.set_tooltip("Min: {:~}\nMax: {:~}".format(self.attribute.minimum(), self.attribute.maximum()))
        return self.input_widget.setMinimum(value.to(self.unit).m)

    def set_maximum(self, value):
        assert isinstance(value, units.Quantity)
        self.set_tooltip("Min: {:~}\nMax: {:~}".format(self.attribute.minimum(), self.attribute.maximum()))
        return self.input_widget.setMaximum(value.to(self.unit).m)

    def on_numerical_value_set(self, value):
        current_font_color = self.font_color
        min_value, max_value = self.input_widget.minimum(), self.input_widget.maximum()
        if min_value == -inf or max_value == inf:
            if value <= min_value or value >= max_value:
                self.font_color = "red"
            else:
                self.font_color = "black"
        else:
            try:
                value_in_percent = ((value - min_value) * 100.) / (max_value - min_value)
            except ZeroDivisionError:
                value_in_percent = 50

            if value_in_percent >= 100 or value_in_percent <= 0:
                self.font_color = "red"
            elif value_in_percent <= 20:
                self.font_color = "purple"
            elif value_in_percent <= 80:
                self.font_color = "black"
            else:
                self.font_color = "blue"

        if current_font_color != self.font_color:
            self.input_widget.setStyleSheet(
                ".CustomDoubleSpinBox, .CustomSpinBox {{background:{}; color:{}}}".format(self.background_color,
                                                                                          self.font_color))

    @pyqtSlot("QString")
    def set_unit(self, unit=None, reset=False):
        # Set widget to new unit
        old_unit, new_unit = self.unit, unit
        old_min, old_max = self.get_minimum(), self.get_maximum()
        self.unit = unit

        if unit is None:
            return
        try:
            new_value = units.Quantity(self.input_widget.value(), old_unit).to(new_unit)
        except AttributeError:
            return
        except DimensionalityError:
            old_unit = new_unit
            new_value = units.Quantity(self.input_widget.value(), new_unit)
            old_min = self.get_minimum()
            old_max = self.get_maximum()

        min_value = max_value = None
        # Recalculate minimum and maximum value to new unit
        try:
            min_value = old_min.to(new_unit)
            if min_value is not None and hasattr(self.input_widget, "setMinimum"):
                self.input_widget.setMinimum(min_value.m)
        except KeyError:
            pass  # Attribute has no minimum
        try:
            max_value = old_max.to(new_unit)
            if max_value is not None and hasattr(self.input_widget, "setMaximum"):
                self.input_widget.setMaximum(max_value.m)
        except KeyError:
            pass  # Attribute has no maximum

        self.set_tooltip("Min: {}\nMax: {}".format(min_value, max_value))

        try:
            if reset:
                decimals = self.config.get("max-decimals", self.attribute[DECIMALS])
                self.input_widget.setDecimals(decimals)
            else:
                decimals_diff = get_decimals_diff(old_unit, new_unit)
                decimals = self.config.get("max-decimals", self.input_widget.decimals() - decimals_diff)
                self.input_widget.setDecimals(decimals)
        except (KeyError, AttributeError):
            pass  # Attribute has no decimals

        # Update displayed value
        self.set_value(new_value)
        self.sig_unit_changed.emit(old_unit, new_unit)



from time import sleep

import pytest

import kamzik3
from kamzik3.devices.deviceSession import DeviceSession


session_config = {'attributes': {('Log directory', 'Value'): './kamzik3/test/server_log',
                                 ('Resource directory', 'Value'): './kamzik3/test/resources',
                                 ('Allow attribute log', 'Value'): False}}
DeviceSession(device_id="Device_Session", config=session_config)
sleep(3)
print("now")
kamzik3.session.stop()



import hashlib
import os
import re
import time

import kamzik3
from kamzik3 import DeviceError
from kamzik3.constants import *
from kamzik3.devices.device import Device
from kamzik3.devices.deviceFileServer import DeviceFileServer
from kamzik3.snippets.snippetDataAccess import compute_MD5
from kamzik3.snippets.snippetsDecorators import expose_method


class DeviceVirtualDirectory(DeviceFileServer):

    def __init__(self, host, port, workers_count=2, drives=None, device_id=None, config=None):
        DeviceFileServer.__init__(self, device_id, host, port, workers_count, config=config)
        self.set_value(ATTR_DRIVES, drives)

    def _init_attributes(self):
        Device._init_attributes(self)
        self.create_attribute(ATTR_DRIVES, set_function=self._set_drives, readonly=True)
        self.create_attribute(ATTR_HOST, default_value=self.host, readonly=True)
        self.create_attribute(ATTR_PORT, default_value=self.port, readonly=True)

    def handle_configuration(self):
        start_at = time.time()
        self._config_commands()
        self._config_attributes()
        self.start()
        self.set_status(STATUS_CONFIGURED)
        self.logger.info(u"Device configuration took {} sec.".format(time.time() - start_at))

    def _set_drives(self, drives):
        """
        Initiate drives for current virtual directory.
        :param drives: dict
        :return: None
        """
        for value in drives.values():
            value["path"] = os.path.normpath(value["path"])
        self.set_value(ATTR_DRIVES, drives)

    def _check_allowed_path(self, drive, path):
        """
        Check if requested path on drive is actually allowed and exists
        :param drive: str
        :param path: str
        :return: bool
        """
        drives = self.get_value(ATTR_DRIVES)

        if drive not in drives:
            raise DeviceError("Drive {} does not exists!".format(drive))

        drive_meta = self.get_value(ATTR_DRIVES)[drive]
        source = drive_meta.get("source", LOCAL_FILESYSTEM)
        if source == LOCAL_FILESYSTEM:
            if not os.path.exists(path):
                raise DeviceError("Path does not exists!")
            for allowed_path in self.get_value(ATTR_DRIVES)[drive]["path"]:
                if allowed_path in path:
                    return True
            else:
                raise DeviceError("Path does not exists!")
        else:
            return True

    @expose_method({"drive": "Drive", "input_dir": "Path to the directory", "file_filter": "Regexp filter"})
    def get_dir_content(self, drive, input_dir, file_filter=None):
        """
        Get all files, dirs and metadata (modtime, filesize) within input_dir and drive.
        You can set file_filter regexp to filter out file names.
        :param drive: str
        :param input_dir:  str
        :param file_filter: regexp str
        :return: str, list, list, list, list
        """
        self._check_allowed_path(drive, input_dir)
        drive_meta = self.get_value(ATTR_DRIVES)[drive]
        source = drive_meta.get("source", LOCAL_FILESYSTEM)
        if file_filter is None:
            file_filter_pattern = re.compile(drive_meta.get("file_filter", ".*"))
        else:
            file_filter_pattern = re.compile(file_filter)

        if source == LOCAL_FILESYSTEM:
            root_dir, dirs, files = next(
                os.walk(input_dir, topdown=True, onerror=None, followlinks=False))
            filtered_files = list(filter(lambda i, p=file_filter_pattern: file_filter_pattern.match(i), files))
            files_meta = []
            dirs_meta = []
            for d_dir in dirs:
                absolute_path = "/".join((root_dir, d_dir))
                dirs_meta.append([0, os.path.getmtime(absolute_path)])
            for d_file in filtered_files:
                absolute_path = "/".join((root_dir, d_file))
                files_meta.append([os.path.getsize(absolute_path), os.path.getmtime(absolute_path)])
            return root_dir, dirs, dirs_meta, filtered_files, files_meta
        else:
            return source.get_dir_content(drive=drive, input_dir=input_dir, file_filter=file_filter)

    @expose_method({"drive": "Drive", "input_dir": "Path to the directory", "file_name": "File name"})
    def get_meta_hash(self, drive, input_dir, file_name):
        """
        Get hash of device_id+drive+file_path this should ensure unique hash for any filepath.
        Return value is used as a path for cached file.
        :param drive: str
        :param input_dir: str
        :param file_name: str
        :return: str
        """
        self._check_allowed_path(drive, input_dir)
        drive_meta = self.get_value(ATTR_DRIVES)[drive]
        source = drive_meta.get("source", LOCAL_FILESYSTEM)
        if source == LOCAL_FILESYSTEM:
            file_path = "/".join((input_dir, file_name))
            with open(file_path, "rb") as file_to_check:
                # read contents of the file
                data = file_to_check.read()
                # pipe contents of the file through
                md5_hash = hashlib.md5(data).hexdigest()
            return compute_MD5("{}{}{}".format(self.device_id, drive, file_path)), os.path.getsize(
                file_path), os.path.getmtime(file_path), md5_hash
        else:
            return source.get_meta_hash(drive=drive, input_dir=input_dir, file_name=file_name)

    @expose_method({"drive": "Drive", "input_dir": "Path to the directory", "file_name": "File name"})
    def get_local_path(self, drive, input_dir, file_name):
        """
        Get local path to either cache or file on local fs
        :param drive: str
        :param input_dir: str
        :param file_name: str
        :return: tuple (bool, str)
        """
        self._check_allowed_path(drive, input_dir)
        drive_meta = self.get_value(ATTR_DRIVES)[drive]
        source = drive_meta.get("source", LOCAL_FILESYSTEM)
        file_path = "/".join((input_dir, file_name))
        if source == LOCAL_FILESYSTEM:
            return True, file_path
        else:
            cached, cache_path = self.is_cached(drive, input_dir, file_name)
            if os.path.exists(cache_path):
                return cached, cache_path
            else:
                return cached, None

    def is_cached(self, drive, input_dir, file_name):
        """
        Check if path on drive from input_dir is cached
        :param drive: str
        :param input_dir: str
        :param file_name: str
        :return: tuple (bool, path_hash)
        """
        cache_directory = kamzik3.session.get_value(ATTR_CACHE_DIRECTORY)
        path_hash, file_size, file_mod_time, file_md5_hash = self.get_meta_hash(drive, input_dir, file_name)
        path_hash = "/".join((cache_directory, path_hash))
        if os.path.exists(path_hash):
            with open(path_hash, "rb") as file_to_check:
                # read contents of the file
                data = file_to_check.read()
                # pipe contents of the file through
                cached_file_hash = hashlib.md5(data).hexdigest()

            if cached_file_hash == file_md5_hash:
                return True, path_hash

        return False, path_hash

    @expose_method({"input_files": "List of tuples (drive, file_path)"})
    def prepare_files(self, input_files):
        """
        Get through all files and get real drive / path combination.
        :param files: list of files
        :return: tuple (list of files, metadata)
        """
        prepared_files = []
        metadata = dict()
        for drive, file_path in input_files:
            if not self.on_fs(drive, file_path):
                drive_meta = self.get_value(ATTR_DRIVES)[drive]
                source = drive_meta["source"]
                source_prepared_files, source_metadata = source.prepare_files(input_files=[(drive, file_path)])
                prepared_files += source_prepared_files
                metadata.update(source_metadata)
            else:
                prepared_files.append((drive, file_path))
        return prepared_files, metadata

    def sync_files(self, requested_files):
        """
        Request files from fileserver
        :param requested_files:
        :return: list
        """
        sync_parts = [INSTRUCTION_SYNC, str(len(requested_files)).encode()]
        for drive, file_path in requested_files:
            sync_parts += [drive.encode(), file_path.encode()]
        self.frontend_connector.send_multipart(sync_parts)
        reply = self.frontend_connector.recv_multipart()
        return reply[1].decode()

    @expose_method({"drive": "Drive", "path": "Path to the file"})
    def on_fs(self, drive, path):
        """
        Check if file path from drive exists on local drive.
        :param drive: str
        :param path: str
        :return: bool
        """
        self._check_allowed_path(drive, path)
        drive_meta = self.get_value(ATTR_DRIVES)[drive]
        source = drive_meta.get("source", LOCAL_FILESYSTEM)
        return source == LOCAL_FILESYSTEM and os.path.exists(path)



from PyQt5.QtCore import QSize, Qt
from PyQt5.QtGui import QCursor, QIcon, QPixmap
from PyQt5.QtWidgets import QPushButton, QSizePolicy


class LockButton(QPushButton):

    def __init__(self, *__args):
        super().__init__(*__args)
        size_policy = QSizePolicy(QSizePolicy.Fixed, QSizePolicy.Fixed)
        size_policy.setHorizontalStretch(0)
        size_policy.setVerticalStretch(0)
        size_policy.setHeightForWidth(self.sizePolicy().hasHeightForWidth())
        self.setSizePolicy(size_policy)
        self.setMaximumSize(QSize(16, 16))
        self.setCursor(QCursor(Qt.OpenHandCursor))
        self.setStyleSheet(u"QPushButton:flat {   border: none; }")
        self.setText(u"")
        icon = QIcon()
        icon.addPixmap(QPixmap(u":/icons/icons/radio_off.png"), QIcon.Normal, QIcon.Off)
        icon.addPixmap(QPixmap(u":/icons/icons/radio_on.png"), QIcon.Normal, QIcon.On)
        self.setIcon(icon)
        self.setIconSize(QSize(16, 16))
        self.setCheckable(True)
        self.setChecked(False)
        self.setDefault(False)
        self.setAutoExclusive(True)
        self.setFlat(True)



import json
import time
from base64 import b64encode, b64decode

import numpy as np

import kamzik3
from kamzik3.devices.device import Device
from kamzik3 import units


class JsonKamzikEncoder(json.JSONEncoder):

    def default(self, o):
        if isinstance(o, units.Quantity):
            return {"pint-unit-quantity": o.to_tuple()}
        elif isinstance(o, np.ndarray):
            return {"numpy-array": o.tolist(), "dtype": o.dtype.name, "shape": o.shape}
        elif isinstance(o, bytes):
            return {"bytes": b64encode(o).decode()}
        elif isinstance(o, Device):
            return {"kamzik3-device": o.device_id}
        return json.JSONEncoder.default(self, o)


def JsonKamzikHook(dct):
    if 'pint-unit-quantity' in dct:
        return units.Quantity.from_tuple(dct['pint-unit-quantity'])
    elif "numpy-array" in dct:
        return np.asarray(dct.get("numpy-array"), dtype=dct.get("dtype"))
    elif "bytes" in dct:
        return b64decode(dct.get("bytes"))
    elif "kamzik3-device" in dct:
        return kamzik3.session.get_device(dct.get("kamzik3-device"))
    return dct



# -*- coding: utf-8 -*-

# Form implementation generated from reading ui file 'deviceDisplayAttributeTemplate.ui'
#
# Created by: PyQt5 UI code generator 5.10.1
#
# WARNING! All changes made in this file will be lost!

from PyQt5 import QtCore, QtGui, QtWidgets

class Ui_Form(object):
    def setupUi(self, Form):
        Form.setObjectName("Form")
        Form.resize(416, 39)
        Form.setStyleSheet("QWidget#widget_holder {background:#fff}")
        self.horizontalLayout = QtWidgets.QHBoxLayout(Form)
        self.horizontalLayout.setContentsMargins(0, 0, 0, 0)
        self.horizontalLayout.setObjectName("horizontalLayout")
        self.widget_holder = QtWidgets.QWidget(Form)
        self.widget_holder.setObjectName("widget_holder")
        self.horizontalLayout_2 = QtWidgets.QHBoxLayout(self.widget_holder)
        self.horizontalLayout_2.setContentsMargins(6, 6, 6, 6)
        self.horizontalLayout_2.setObjectName("horizontalLayout_2")
        self.label_device_name = QtWidgets.QLabel(self.widget_holder)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Fixed, QtWidgets.QSizePolicy.Preferred)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.label_device_name.sizePolicy().hasHeightForWidth())
        self.label_device_name.setSizePolicy(sizePolicy)
        self.label_device_name.setMinimumSize(QtCore.QSize(150, 0))
        self.label_device_name.setMaximumSize(QtCore.QSize(16777215, 16777215))
        font = QtGui.QFont()
        font.setFamily("Sans Serif")
        font.setPointSize(9)
        font.setBold(True)
        font.setWeight(75)
        self.label_device_name.setFont(font)
        self.label_device_name.setStyleSheet("")
        self.label_device_name.setScaledContents(False)
        self.label_device_name.setObjectName("label_device_name")
        self.horizontalLayout_2.addWidget(self.label_device_name)
        self.label_status = QtWidgets.QLabel(self.widget_holder)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Fixed, QtWidgets.QSizePolicy.Preferred)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.label_status.sizePolicy().hasHeightForWidth())
        self.label_status.setSizePolicy(sizePolicy)
        self.label_status.setMinimumSize(QtCore.QSize(100, 0))
        self.label_status.setMaximumSize(QtCore.QSize(16777215, 16777215))
        font = QtGui.QFont()
        font.setPointSize(10)
        font.setBold(True)
        font.setWeight(75)
        font.setKerning(True)
        self.label_status.setFont(font)
        self.label_status.setStyleSheet("QLabel {background:#68d532}")
        self.label_status.setAlignment(QtCore.Qt.AlignCenter)
        self.label_status.setObjectName("label_status")
        self.horizontalLayout_2.addWidget(self.label_status)
        self.input_not_ready_placeholder = QtWidgets.QSpinBox(self.widget_holder)
        self.input_not_ready_placeholder.setEnabled(False)
        self.input_not_ready_placeholder.setReadOnly(True)
        self.input_not_ready_placeholder.setObjectName("input_not_ready_placeholder")
        self.horizontalLayout_2.addWidget(self.input_not_ready_placeholder)
        self.button_debug = QtWidgets.QPushButton(self.widget_holder)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Fixed, QtWidgets.QSizePolicy.Fixed)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.button_debug.sizePolicy().hasHeightForWidth())
        self.button_debug.setSizePolicy(sizePolicy)
        self.button_debug.setText("")
        icon = QtGui.QIcon()
        icon.addPixmap(QtGui.QPixmap(":/icons/icons/settings.png"), QtGui.QIcon.Normal, QtGui.QIcon.Off)
        self.button_debug.setIcon(icon)
        self.button_debug.setIconSize(QtCore.QSize(19, 19))
        self.button_debug.setObjectName("button_debug")
        self.horizontalLayout_2.addWidget(self.button_debug)
        self.horizontalLayout.addWidget(self.widget_holder)

        self.retranslateUi(Form)
        self.button_debug.clicked.connect(Form.show_debug_widget)
        QtCore.QMetaObject.connectSlotsByName(Form)

    def retranslateUi(self, Form):
        _translate = QtCore.QCoreApplication.translate
        Form.setWindowTitle(_translate("Form", "Form"))
        self.label_device_name.setText(_translate("Form", "ID"))
        self.label_status.setText(_translate("Form", "STATUS"))
        self.input_not_ready_placeholder.setSpecialValueText(_translate("Form", "Not ready"))
        self.button_debug.setToolTip(_translate("Form", "Settings"))

from kamzik3 import resource_kamzik3_rc



import pytest

import kamzik3
from kamzik3.constants import *
from kamzik3.devices.device import Device


@pytest.fixture
def device():
    yield Device(device_id="test_device", config=None)


def test_device_creation(device):
    assert device.device_id == "test_device"
    assert device.session is kamzik3.session
    assert device.get_value(ATTR_ID) == "test_device"


@pytest.mark.parametrize("status", STATUSES_PRIORITY)
def test_device_status(device, status):
    device.set_status(status)
    assert device.get_value(ATTR_STATUS) == status



import re
import time
from decimal import DivisionByZero
from math import ceil

import bidict
import numpy as np
import serial
from bidict import frozenbidict, MutableBidict
from numpy import around, sign, sqrt
from pint.errors import UndefinedUnitError

import kamzik3
from kamzik3 import DeviceUnitError, DeviceError
from kamzik3.constants import *
from kamzik3.devices.deviceChannel import DeviceChannel
from kamzik3.devices.devicePort import DevicePort
from kamzik3.devices.deviceSocket import DeviceSocket
from kamzik3.snippets.snippetsDecorators import expose_method
from kamzik3.snippets.snippetsUnits import device_units

TRIG_OUTPUT_STEP = "Trigger step"
TRIG_OUTPUT_AXIS = "Trigger axis"
TRIG_OUTPUT_MODE = "Trigger mode"
TRIG_OUTPUT_MIN_THRESHOLD = "Trigger min threshold"
TRIG_OUTPUT_MAX_THRESHOLD = "Trigger max threshold"
TRIG_POLARITY = "Trigger polarity"
TRIG_START = "Trigger start"
TRIG_STOP = "Trigger stop"

trigger_output_config = frozenbidict({
    TRIG_OUTPUT_STEP: 1,
    TRIG_OUTPUT_AXIS: 2,
    TRIG_OUTPUT_MODE: 3,
    TRIG_OUTPUT_MIN_THRESHOLD: 5,
    TRIG_OUTPUT_MAX_THRESHOLD: 6,
    TRIG_POLARITY: 7,
    TRIG_START: 8,
    TRIG_STOP: 9,
})

TRIG_MODE_POSITION_DISTANCE = "Position distance"
TRIG_MODE_ON_TARGET = "On target"
TRIG_MODE_WITHIN_THRESHOLD = "Within thresholds"
TRIG_MODE_GENERATOR = "Generator"
trigger_output_mode = frozenbidict({
    TRIG_MODE_POSITION_DISTANCE: 0,
    TRIG_MODE_ON_TARGET: 2,
    TRIG_MODE_WITHIN_THRESHOLD: 3,
    TRIG_MODE_GENERATOR: 4,
})

INTERPOLATION_NONE = "No interpolation"
INTERPOLATION_LINE = "Line interpolation"
data_interpolation = frozenbidict({
    INTERPOLATION_NONE: 0,
    INTERPOLATION_LINE: 1,
})

WAVE_TYPE_PNT = "PNT"  # User defined curve
WAVE_TYPE_SINP = "SIN_P"  # Inverted cosine curve
WAVE_TYPE_RAMP = "RAMP"  # Ramp curve
WAVE_TYPE_LIN = "LIN"  # Inverted cosine curve
WAVE_TYPE_APPEND = "&"
WAVE_TYPE_START = "X"

SERVO_UPDATE_TIME_PARAM = "0xe000200"

REC_TARGET_POSITION = 1
REC_CURRENT_POSITION = 2
REC_ERROR = 3
REC_DDL_OUTPUT = 13
REC_OPEN_LOOP = 14
REC_CONTROL_OUTPUT = 15
REC_SLOWED_TARGET = 22


class DevicePi(DeviceSocket):
    terminator = b"\n"
    error_pattern = re.compile('^([-0-9]+)$')
    response_timeout = 3000
    PRE_SCAN_POINTS = 10
    MAX_SPEED_UMPMS = 10.  # speed initially set in controller 10 mm/sec
    # to reduce communication time, we can pack up to 27 points of a WaveTable into a single "WAV ... PNT" command
    N_WAV_POINTS_PER_CMD = 27
    # to reduce communication time, we can pack up to 10 trigger actions into a single "TWS" command
    N_TRIG_PER_CMD = 10
    CTR_NAME = "Dev1/ctr0"
    CTR_MAX_SEC = 50.  # in this case, we can count with 80MHz clock (-> highest precision)
    error_description = {
        0: "PI_CNTR_NO_ERROR",
        1: "PI_CNTR_PARAM_SYNTAX",
        2: "PI_CNTR_UNKNOWN_COMMAND",
        3: "PI_CNTR_COMMAND_TOO_LONG",
        4: "PI_CNTR_SCAN_ERROR",
        5: "PI_CNTR_MOVE_WITHOUT_REF_OR_NO_SERVO",
        6: "PI_CNTR_INVALID_SGA_PARAM",
        7: "PI_CNTR_POS_OUT_OF_LIMITS",
        8: "PI_CNTR_VEL_OUT_OF_LIMITS",
        9: "PI_CNTR_SET_PIVOT_NOT_POSSIBLE",
        10: "PI_CNTR_STOP",
        11: "PI_CNTR_SST_OR_SCAN_RANGE",
        12: "PI_CNTR_INVALID_SCAN_AXES",
        13: "PI_CNTR_INVALID_NAV_PARAM",
        14: "PI_CNTR_INVALID_ANALOG_INPUT",
        15: "PI_CNTR_INVALID_AXIS_IDENTIFIER",
        16: "PI_CNTR_INVALID_STAGE_NAME",
        17: "PI_CNTR_PARAM_OUT_OF_RANGE",
        18: "PI_CNTR_INVALID_MACRO_NAME",
        19: "PI_CNTR_MACRO_RECORD",
        20: "PI_CNTR_MACRO_NOT_FOUND",
        21: "PI_CNTR_AXIS_HAS_NO_BRAKE",
        22: "PI_CNTR_DOUBLE_AXIS",
        23: "PI_CNTR_ILLEGAL_AXIS",
        24: "PI_CNTR_PARAM_NR",
        25: "PI_CNTR_INVALID_REAL_NR",
        26: "PI_CNTR_MISSING_PARAM",
        27: "PI_CNTR_SOFT_LIMIT_OUT_OF_RANGE",
        28: "PI_CNTR_NO_MANUAL_PAD",
        29: "PI_CNTR_NO_JUMP",
        30: "PI_CNTR_INVALID_JUMP",
        31: "PI_CNTR_AXIS_HAS_NO_REFERENCE",
        32: "PI_CNTR_STAGE_HAS_NO_LIM_SWITCH",
        33: "PI_CNTR_NO_RELAY_CARD",
        34: "PI_CNTR_CMD_NOT_ALLOWED_FOR_STAGE",
        35: "PI_CNTR_NO_DIGITAL_INPUT",
        36: "PI_CNTR_NO_DIGITAL_OUTPUT",
        37: "PI_CNTR_NO_MCM",
        38: "PI_CNTR_INVALID_MCM",
        39: "PI_CNTR_INVALID_CNTR_NUMBER",
        40: "PI_CNTR_NO_JOYSTICK_CONNECTED",
        41: "PI_CNTR_INVALID_EGE_AXIS",
        42: "PI_CNTR_SLAVE_POSITION_OUT_OF_RANGE",
        43: "PI_CNTR_COMMAND_EGE_SLAVE",
        44: "PI_CNTR_JOYSTICK_CALIBRATION_FAILED",
        45: "PI_CNTR_REFERENCING_FAILED",
        46: "PI_CNTR_OPM_MISSING",
        47: "PI_CNTR_OPM_NOT_INITIALIZED",
        48: "PI_CNTR_OPM_COM_ERROR",
        49: "PI_CNTR_MOVE_TO_LIMIT_SWITCH_FAILED",
        50: "PI_CNTR_REF_WITH_REF_DISABLED",
        51: "PI_CNTR_AXIS_UNDER_JOYSTICK_CONTROL",
        52: "PI_CNTR_COMMUNICATION_ERROR",
        53: "PI_CNTR_DYNAMIC_MOVE_IN_PROCESS",
        54: "PI_CNTR_UNKNOWN_PARAMETER",
        55: "PI_CNTR_NO_REP_RECORDED",
        56: "PI_CNTR_INVALID_PASSWORD",
        57: "PI_CNTR_INVALID_RECORDER_CHAN",
        58: "PI_CNTR_INVALID_RECORDER_SRC_OPT",
        59: "PI_CNTR_INVALID_RECORDER_SRC_CHAN",
        60: "PI_CNTR_PARAM_PROTECTION",
        61: "PI_CNTR_AUTOZERO_RUNNING",
        62: "PI_CNTR_NO_LINEAR_AXIS",
        63: "PI_CNTR_INIT_RUNNING",
        64: "PI_CNTR_READ_ONLY_PARAMETER",
        65: "PI_CNTR_PAM_NOT_FOUND",
        66: "PI_CNTR_VOL_OUT_OF_LIMITS",
        67: "PI_CNTR_WAVE_TOO_LARGE",
        68: "PI_CNTR_NOT_ENOUGH_DDL_MEMORY",
        69: "PI_CNTR_DDL_TIME_DELAY_TOO_LARGE",
        70: "PI_CNTR_DIFFERENT_ARRAY_LENGTH",
        71: "PI_CNTR_GEN_SINGLE_MODE_RESTART",
        72: "PI_CNTR_ANALOG_TARGET_ACTIVE",
        73: "PI_CNTR_WAVE_GENERATOR_ACTIVE",
        74: "PI_CNTR_AUTOZERO_DISABLED",
        75: "PI_CNTR_NO_WAVE_SELECTED",
        76: "PI_CNTR_IF_BUFFER_OVERRUN",
        77: "PI_CNTR_NOT_ENOUGH_RECORDED_DATA",
        78: "PI_CNTR_TABLE_DEACTIVATED",
        79: "PI_CNTR_OPENLOOP_VALUE_SET_WHEN_SERVO_ON",
        80: "PI_CNTR_RAM_ERROR",
        81: "PI_CNTR_MACRO_UNKNOWN_COMMAND",
        82: "PI_CNTR_MACRO_PC_ERROR",
        83: "PI_CNTR_JOYSTICK_ACTIVE",
        84: "PI_CNTR_MOTOR_IS_OFF",
        85: "PI_CNTR_ONLY_IN_MACRO",
        86: "PI_CNTR_JOYSTICK_UNKNOWN_AXIS",
        87: "PI_CNTR_JOYSTICK_UNKNOWN_ID",
        88: "PI_CNTR_REF_MODE_IS_ON",
        89: "PI_CNTR_NOT_ALLOWED_IN_CURRENT_MOTION_MODE",
        90: "PI_CNTR_DIO_AND_TRACING_NOT_POSSIBLE",
        91: "PI_CNTR_COLLISION",
        92: "PI_CNTR_SLAVE_NOT_FAST_ENOUGH",
        93: "PI_CNTR_CMD_NOT_ALLOWED_WHILE_AXIS_IN_MOTION This command is not allowed",
        94: "PI_CNTR_OPEN_LOOP_JOYSTICK_ENABLED",
        95: "PI_CNTR_INVALID_SERVO_STATE_FOR_PARAMETER",
        96: "PI_CNTR_UNKNOWN_STAGE_NAME",
        100: "PI_LABVIEW_ERROR",
        200: "PI_CNTR_NO_AXIS",
        201: "PI_CNTR_NO_AXIS_PARAM_FILE",
        202: "PI_CNTR_INVALID_AXIS_PARAM_FILE",
        203: "PI_CNTR_NO_AXIS_PARAM_BACKUP",
        204: "PI_CNTR_RESERVED_204",
        205: "PI_CNTR_SMO_WITH_SERVO_ON",
        206: "PI_CNTR_UUDECODE_INCOMPLETE_HEADER",
        207: "PI_CNTR_UUDECODE_NOTHING_TO_DECODE",
        208: "PI_CNTR_UUDECODE_ILLEGAL_FORMAT",
        209: "PI_CNTR_CRC32_ERROR",
        210: "PI_CNTR_ILLEGAL_FILENAME",
        211: "PI_CNTR_FILE_NOT_FOUND",
        212: "PI_CNTR_FILE_WRITE_ERROR",
        213: "PI_CNTR_DTR_HINDERS_VELOCITY_CHANGE",
        214: "PI_CNTR_POSITION_UNKNOWN",
        215: "PI_CNTR_CONN_POSSIBLY_BROKEN",
        216: "PI_CNTR_ON_LIMIT_SWITCH",
        217: "PI_CNTR_UNEXPECTED_STRUT_STOP",
        218: "PI_CNTR_POSITION_BASED_ON_ESTIMATION",
        219: "PI_CNTR_POSITION_BASED_ON_INTERPOLATION",
        230: "PI_CNTR_INVALID_HANDLE",
        231: "PI_CNTR_NO_BIOS_FOUND",
        232: "PI_CNTR_SAVE_SYS_CFG_FAILED",
        233: "PI_CNTR_LOAD_SYS_CFG_FAILED",
        301: "PI_CNTR_SEND_BUFFER_OVERFLOW",
        302: "PI_CNTR_VOLTAGE_OUT_OF_LIMITS",
        303: "PI_CNTR_OPEN_LOOP_MOTION_SET_WHEN_SERVO_ON Open-loop motion attempted",
        304: "PI_CNTR_RECEIVING_BUFFER_OVERFLOW",
        305: "PI_CNTR_EEPROM_ERROR",
        306: "PI_CNTR_I2C_ERROR",
        307: "PI_CNTR_RECEIVING_TIMEOUT",
        308: "PI_CNTR_TIMEOUT",
        309: "PI_CNTR_MACRO_OUT_OF_SPACE",
        310: "PI_CNTR_EUI_OLDVERSION_CFGDATA",
        311: "PI_CNTR_EUI_INVALID_CFGDATA",
        333: "PI_CNTR_HARDWARE_ERROR",
        400: "PI_CNTR_WAV_INDEX_ERROR",
        401: "PI_CNTR_WAV_NOT_DEFINED",
        402: "PI_CNTR_WAV_TYPE_NOT_SUPPORTED",
        403: "PI_CNTR_WAV_LENGTH_EXCEEDS_LIMIT",
        404: "PI_CNTR_WAV_PARAMETER_NR",
        405: "PI_CNTR_WAV_PARAMETER_OUT_OF_LIMIT",
        406: "PI_CNTR_WGO_BIT_NOT_SUPPORTED",
        500: "PI_CNTR_EMERGENCY_STOP_BUTTON_ACTIVATED",
        501: "PI_CNTR_EMERGENCY_STOP_BUTTON_WAS_ACTIVATED",
        502: "PI_CNTR_REDUNDANCY_LIMIT_EXCEEDED",
        503: "PI_CNTR_COLLISION_SWITCH_ACTIVATED",
        504: "PI_CNTR_FOLLOWING_ERROR",
        555: "PI_CNTR_UNKNOWN_ERROR",
        601: "PI_CNTR_NOT_ENOUGH_MEMORY",
        602: "PI_CNTR_HW_VOLTAGE_ERROR",
        603: "PI_CNTR_HW_TEMPERATURE_ERROR",
        604: "PI_CNTR_POSITION_ERROR_TOO_HIGH",
        606: "PI_CNTR_INPUT_OUT_OF_RANGE",
        1000: "PI_CNTR_TOO_MANY_NESTED_MACROS",
        1001: "PI_CNTR_MACRO_ALREADY_DEFINED",
        1002: "PI_CNTR_NO_MACRO_RECORDING",
        1003: "PI_CNTR_INVALID_MAC_PARAM",
        1004: "PI_CNTR_RESERVED_1004",
        1005: "PI_CNTR_CONTROLLER_BUSY",
        1006: "PI_CNTR_INVALID_IDENTIFIER",
        1007: "PI_CNTR_UNKNOWN_VARIABLE_OR_ARGUMENT",
        1008: "PI_CNTR_RUNNING_MACRO",
        1009: "PI_CNTR_MACRO_INVALID_OPERATOR",
        1063: "PI_CNTR_EXT_PROFILE_UNALLOWED_CMD",
        1064: "PI_CNTR_EXT_PROFILE_EXPECTING_MOTION_ERROR",
        1065: "PI_CNTR_PROFILE_ACTIVE",
        1066: "PI_CNTR_PROFILE_INDEX_OUT_OF_RANGE",
        1071: "PI_CNTR_PROFILE_OUT_OF_MEMORY",
        1072: "PI_CNTR_PROFILE_WRONG_CLUSTER",
        1073: "PI_CNTR_PROFILE_UNKNOWN_CLUSTER_IDENTIFIER",
        2000: "PI_CNTR_ALREADY_HAS_SERIAL_NUMBER",
        4000: "PI_CNTR_SECTOR_ERASE_FAILED",
        4001: "PI_CNTR_FLASH_PROGRAM_FAILED",
        4002: "PI_CNTR_FLASH_READ_FAILED",
        4003: "PI_CNTR_HW_MATCHCODE_ERROR",
        4004: "PI_CNTR_FW_MATCHCODE_ERROR",
        4005: "PI_CNTR_HW_VERSION_ERROR",
        4006: "PI_CNTR_FW_VERSION_ERROR",
        4007: "PI_CNTR_FW_UPDATE_ERROR",
        4008: "PI_CNTR_FW_CRC_PAR_ERROR",
        4009: "PI_CNTR_FW_CRC_FW_ERROR",
        5000: "PI_CNTR_INVALID_PCC_SCAN_DATA",
        5001: "PI_CNTR_PCC_SCAN_RUNNING",
        5002: "PI_CNTR_INVALID_PCC_AXIS",
        5003: "PI_CNTR_PCC_SCAN_OUT_OF_RANGE",
        5004: "PI_CNTR_PCC_TYPE_NOT_EXISTING",
        5005: "PI_CNTR_PCC_PAM_ERROR",
        5006: "PI_CNTR_PCC_TABLE_ARRAY_TOO_LARGE",
        5100: "PI_CNTR_NEXLINE_ERROR",
        5101: "PI_CNTR_CHANNEL_ALREADY_USED",
        5102: "PI_CNTR_NEXLINE_TABLE_TOO_SMALL",
        5103: "PI_CNTR_RNP_WITH_SERVO_ON",
        5104: "PI_CNTR_RNP_NEEDED",
        5200: "PI_CNTR_AXIS_NOT_CONFIGURED",
    }

    def __init__(self, host, port=50000, device_id=None, config=None):
        self.channel_mapping = MutableBidict()
        self.valid_commands = []
        super().__init__(host, port, device_id, config)

    def _init_attributes(self):
        super(DevicePi, self)._init_attributes()
        self.create_attribute(ATTR_SERIAL_NUMBER, readonly=True)
        self.create_attribute(ATTR_CHANNELS, readonly=True, default_type=np.uint16)
        self.create_attribute(ATTR_SYNTAX_VERSION, readonly=True, default_type=np.float16, decimals=3)
        self.create_attribute(ATTR_POSITIONS, readonly=True)
        self.create_attribute(ATTR_TRIGGER_OUTPUTS_COUNT, readonly=True, default_type=np.uint16)
        self.create_attribute(ATTR_WAVE_GEN_COUNT, readonly=True, default_type=np.uint16)
        self.create_attribute(ATTR_SERVO_UPDATE_TIME, readonly=True, default_type=np.float64, decimals=12, unit="ms")

    def valid_command_format(self, command):
        if command is not None and command[-1] == "\n":
            return True
        else:
            return False

    def handle_configuration(self):
        start_at = time.time()

        def _finish_configuration(*_, **__):
            self._config_commands()
            self._config_attributes()
            self.start_polling()
            self.set_status(STATUS_CONFIGURED)
            self.logger.info(u"Device configuration took {} sec.".format(time.time() - start_at))

        self.command("ERR?\n")
        # Get PI serial number
        self.command(u"IDN?\n")
        # Get PI syntax version
        self.command(u"CSV?\n")
        # Get available commands
        self.command(u"HLP?\n")
        # Get Channels configuration
        self.get_memory_parameters([SERVO_UPDATE_TIME_PARAM])
        self.get_configuration_of_trigger_outputs(callback=self._set_trigger_output)
        self.get_wave_generator_mode(callback=self._set_wave_generators)
        self.command(u"SAI?\n", callback=_finish_configuration)

    def _set_trigger_output(self, command, output):
        output = "".join(output).split(" \n")
        trigger_output_count = 0
        for line in output:
            trigger_output, value = line.split(" ")
            attribute, value = value.split("=")

            attribute = trigger_output_config.inv.get(int(attribute), None)
            if attribute is not None:
                trigger_output = int(trigger_output) - 1
                trigger_group_name = "Trigger output {}".format(trigger_output)

                if self.attributes.get(trigger_group_name) is None:
                    self.attributes[trigger_group_name] = {}
                    trigger_output_count += 1

                self.create_attribute(ATTR_TRIGGER_LINE_DATA, trigger_group_name, readonly=True)
                set_attribute = lambda value, parameter=attribute, trigger_output=trigger_output: self.set_configuration_of_trigger_output(trigger_output, parameter, value)
                if attribute == TRIG_OUTPUT_AXIS:
                    set_attribute = lambda value, parameter=attribute, trigger_output=trigger_output: self.set_configuration_of_trigger_output(trigger_output, parameter, int(value) + 1)
                    self.create_attribute(attribute, trigger_group_name, default_value=int(value) - 1, default_type=np.uint16,
                                                                               set_function=set_attribute)
                elif attribute == TRIG_OUTPUT_MODE:
                    self.create_attribute(attribute, trigger_group_name, default_value=trigger_output_mode.inv.get(int(value)),
                                                                               default_type=list(
                                                                                   trigger_output_mode.keys()),
                                                                               set_function=set_attribute)
                else:
                    self.create_attribute(attribute, trigger_group_name, default_value=float(value), default_type=np.float64,
                                                                               decimals=12, set_function=set_attribute)

        self.set_value(ATTR_TRIGGER_OUTPUTS_COUNT, trigger_output_count)

    def _set_wave_generators(self, command, output):
        output = "".join(output).split(" \n")
        for line in output:
            generator_id, mode = line.split("=")
            generator_id = int(generator_id) - 1
            mode = int(mode)
            generator_group_name = "Wave generator {}".format(generator_id)
            self.create_attribute(ATTR_STATUS, generator_group_name, default_value=STATUS_IDLE, readonly=True)
            self.create_attribute(ATTR_WAVETABLE_DATA, generator_group_name, readonly=True)
            self.create_attribute(ATTR_RECORD_TABLE_DATA, generator_group_name, readonly=True)
            self.create_attribute(ATTR_WAVETABLE_LENGTH, generator_group_name, readonly=True, default_type=np.uint32)
            set_function = lambda offset, gen_id=generator_id: self.set_wave_gen_output_offset(gen_id, offset)
            self.create_attribute(ATTR_WAVETABLE_OFFSET, generator_group_name, readonly=False, decimals=12,
                                                                                     default_type=np.float64,
                                                                                     set_function=set_function)
            self.create_attribute(ATTR_WAVE_TABLE, generator_group_name, readonly=True, default_type=np.uint32)

            set_function = lambda table_rate, gen_id=generator_id: self.set_wave_generator_rate_and_interpolation_type(
                gen_id, wave_table_rate=table_rate)
            self.create_attribute(ATTR_WAVETABLE_RATE, generator_group_name, default_type=np.uint32, set_function=set_function)

            set_function = lambda interpolation_type, gen_id=generator_id: self.set_wave_generator_rate_and_interpolation_type(gen_id, interpolation_type=interpolation_type)
            self.create_attribute(ATTR_WAVETABLE_INTERPOLATION, generator_group_name, default_type=list(data_interpolation.keys()), set_function=set_function)

        self.set_value(ATTR_WAVE_GEN_COUNT, len(output))
        self.get_wave_generator_rate_and_interpolation_type()
        self.get_wave_gen_output_offset()
        self.get_connected_table_to_generator()

    def found_terminator(self):
        if len(self.buffer) == 0:
            return
        elif self.buffer[-1][-1] == " ":
            self.buffer[-1] += self.terminator.decode()
        else:
            super().found_terminator()

    def collect_incoming_data(self, data):
        self.buffer.append(data.decode("latin-1"))

    def command(self, command, callback=None, with_token=False, returning=True):
        # Send requested format
        out = super().command(command, callback=callback, with_token=with_token, returning=True)
        # Send command to check for possible error immediately after
        if not returning:
            super().command("ERR?\n", callback=None, with_token=with_token, returning=False)
        return out

    def handle_readout(self, readout_buffer):
        command, output, callback, token = super(DevicePi, self).handle_readout(readout_buffer)
        is_error_message = self.error_pattern.match(output)
        if is_error_message is not None:
            error_code = int(is_error_message.group(0))
            if command == "ERR?\n" and error_code != 0:
                readout_buffer = self.error_description.get(error_code, "Unknown error")
                self.handle_command_error(command, readout_buffer)
                command = RESPONSE_ERROR
            elif command not in ("#5\n", "#9\n", "STP\n") and error_code != 0:
                command = RESPONSE_ERROR
                readout_buffer = self.error_description.get(error_code, "Unknown error")
                self.handle_command_error(command, readout_buffer)

        if command == "IDN?\n":
            self.set_raw_value(ATTR_SERIAL_NUMBER, "{}".format(output))
        elif command == "SAI?\n":
            for channel_id in output.split(" \n"):
                try:
                    self.channel_mapping[channel_id] = len(self.channel_mapping)
                except bidict._exc.KeyAndValueDuplicationError:
                    continue
            self.set_raw_value(ATTR_CHANNELS, len(self.channel_mapping))
        elif command == "CSV?\n":
            self.set_raw_value(ATTR_SYNTAX_VERSION, float(output))
        elif command == "HLP?\n":
            for valid_command in output.split("\n")[1:-1]:
                cmd, desc = valid_command.split(" - ")
                self.valid_commands.append(cmd)
        elif command == "#5\n":
            stat = int(output)
            for channel in range(0, self.get_value(ATTR_CHANNELS)):
                self.notify((channel, ATTR_MOVING), bool(stat & (1 << channel)))
        elif command == "#9\n":
            stat = int(output)
            for generator_id in range(0, self.get_value(ATTR_WAVE_GEN_COUNT)):
                state = bool(stat & (1 << generator_id))
                if not state and self.get_value(("Wave generator {}".format(generator_id), ATTR_STATUS)) == STATUS_BUSY:
                    self.get_recorder_data(generator_id)
                self.set_raw_value(("Wave generator {}".format(generator_id), ATTR_STATUS),
                               STATUS_BUSY if state else STATUS_IDLE)
                trigger_axis = self.get_value(("Trigger output {}".format(generator_id), TRIG_OUTPUT_AXIS))
                self.notify((trigger_axis, ATTR_SCANNING), state)
            if stat == 0:
                self.set_status(STATUS_IDLE)
            else:
                self.set_status(STATUS_BUSY)
        elif "PUN?" in command:
            channel, unit = output.split("=")
            self.notify((self.channel_mapping[channel], ATTR_ENCODER_BASE_UNIT), unit)
        elif "POS?" in command:
            positions = output.split(" \n")
            positions_list = []
            for position in positions:
                channel, position = position.split("=")
                positions_list.append(position)
                self.notify((self.channel_mapping[channel], ATTR_POSITION), float(position))
            self.set_raw_value(ATTR_POSITIONS, " ".join(positions_list))
        elif "FRF?" in command:
            channel, flag = output.split("=")
            self.notify((self.channel_mapping[channel], ATTR_REFERENCED), bool(int(flag)))
        elif "SVO?" in command:
            channel, flag = output.split("=")
            self.notify((self.channel_mapping[channel], ATTR_SERVO_STATE), bool(int(flag)))
        elif "TMN?" in command:
            channel, position = output.split("=")
            self.notify((self.channel_mapping[channel], ATTR_POSITION_LOWER_LIMIT), float(position))
        elif "TMX?" in command:
            channel, position = output.split("=")
            self.notify((self.channel_mapping[channel], ATTR_POSITION_UPPER_LIMIT), float(position))
        elif "VEL?" in command:
            channel, velocity = output.split("=")
            self.notify((self.channel_mapping[channel], ATTR_VELOCITY), float(velocity))
        elif "WOS?" in command:
            lines = output.split(" \n")
            for line in lines:
                generator_id, generator_offset = line.split("=")
                self.set_raw_value(["Wave generator {}".format(int(generator_id) - 1), ATTR_WAVETABLE_OFFSET],
                               float(generator_offset))
        elif "WSL?" in command:
            lines = output.split(" \n")
            for line in lines:
                generator_id, wave_table_id = line.split("=")
                self.set_raw_value(["Wave generator {}".format(int(generator_id) - 1), ATTR_WAVE_TABLE], int(wave_table_id))
        elif "WTR?" in command:
            lines = output.split(" \n")
            for line in lines:
                generator_id, rate_interpolation = line.split("=")
                rate, interpolation = rate_interpolation.split(" ")
                interpolation = data_interpolation.inv.get(int(interpolation))
                if int(generator_id) == 0:
                    for generator_id in range(self.get_value(ATTR_WAVE_GEN_COUNT)):
                        self.set_raw_value(["Wave generator {}".format(generator_id), ATTR_WAVETABLE_RATE],
                                       int(rate))
                        self.set_raw_value(["Wave generator {}".format(generator_id), ATTR_WAVETABLE_INTERPOLATION],
                                       interpolation)
                else:
                    self.set_raw_value(["Wave generator {}".format(int(generator_id) - 1), ATTR_WAVETABLE_RATE],
                                   int(rate))
                    self.set_raw_value(["Wave generator {}".format(int(generator_id) - 1), ATTR_WAVETABLE_INTERPOLATION],
                                   interpolation)
        elif "SEP?" in command:
            lines = output.split(" \n")
            for line in lines:
                parameter, value = line.split("=")
                _, parameter = parameter.split(" ")

                if parameter == SERVO_UPDATE_TIME_PARAM:
                    self.set_raw_value(ATTR_SERVO_UPDATE_TIME, float(value) * 1e3)
        elif "GWD?" in command:
            lines = output.split(" \n")
            wave_table_id = int(lines[5][-1])
            wave_table_data = ", ".join(lines[7:])
            generator_id = wave_table_id - 1
            self.set_raw_value(["Wave generator {}".format(generator_id), ATTR_WAVETABLE_DATA],
                           wave_table_data)
        elif "TWS?" in command:
            lines = output.split(" \n")
            cmd, start, end, trigger_output = command.split(" ")
            trigger_group_name = "Trigger output {}".format(int(trigger_output) - 1)
            triggers_level = ", ".join(lines[6:])
            self.set_raw_value([trigger_group_name, ATTR_TRIGGER_LINE_DATA], triggers_level)
        elif "DRR?" in command:
            lines = output.split(" \n")
            wave_table_id = int(lines[5][-1])
            recorded_data = ", ".join(lines[7:])
            generator_id = wave_table_id - 1
            self.set_raw_value(["Wave generator {}".format(generator_id), ATTR_RECORD_TABLE_DATA],
                           recorded_data)

            trigger_group_name = "Trigger output {}".format(generator_id)
            channel = self.get_value([trigger_group_name, "Trigger axis"])
            self.notify((channel, ATTR_RECORD_TABLE_DATA), recorded_data)
        if callback is not None:
            self.handle_readout_callback(callback, command, readout_buffer)
        if token:
            kamzik3.session.publisher.push_message(self.device_id, [command, readout_buffer], token)
            for observer in self._observers[:]:
                if isinstance(observer, DevicePiChannel):
                    kamzik3.session.publisher.push_message(observer.device_id, [command, readout_buffer], token)

    def start_polling(self):
        super(DevicePi, self).start_polling()
        self.poll_command("POS?\n", 200)
        self.poll_command("#5\n", 200)
        self.poll_command("#9\n", 300)

    def get_memory_parameters(self, list_of_parameters, callback=None):
        assert isinstance(list_of_parameters, list)
        attributes = " ".join(
            [" ".join(c) for c in zip([str(b + 1) for b in range(len(list_of_parameters))], list_of_parameters)])
        return self.command("SEP? {}\n".format(attributes), with_token=True, callback=callback)

    def clear_wave_table(self, wave_table_id):
        """
        Clears the content of the given wave table
        As long as a wave generator is running, it is not
        possible to clear the connected wave table.
        :param wave_table_id: int
        :return:
        """
        self.logger.info("Clearing wave table generator {}".format(wave_table_id))
        return self.command("WCL {}\n".format(wave_table_id + 1), with_token=True, returning=False)

    def clear_wave_table_triggers(self):
        """
        Clears all output trigger settings for the wave
        generators (the settings made with TWS by
        switching the signal state for all points to "low".
        :return:
        """
        self.logger.info("Clearing all wave table triggers")
        return self.command("TWC\n", with_token=True, returning=False)

    def get_wave_generator_rate_and_interpolation_type(self, callback=None):
        """
        Gets the current wave generator table rate, i.e. the
        number of servo-loop cycles used by the wave
        generator to output one waveform point (Wave
        Generator Table Rate parameter value in volatile
        memory (ID 0x13000109)). Gets also the interpolation
        type used with table rate values > 1
        """
        return self.command("WTR?\n", with_token=True, callback=callback)

    def set_wave_generator_rate_and_interpolation_type(self, wave_generator_id=0, wave_table_rate=None,
                                                       interpolation_type=None,
                                                       callback=None):
        """
        Set wave generator interpolation type.
        :param wave_generator_id: int
        :param wave_table_rate: int
        :param interpolation_type: int
        :param callback:
        :type interpolation_type: 0 = no interpolation, 1 = straight line (default)
        :return:
        """
        if wave_table_rate is None:
            wave_table_rate = self.get_value(["Wave generator {}".format(wave_generator_id), ATTR_WAVETABLE_RATE])
        if interpolation_type is None:
            interpolation_type = self.get_value(
                ["Wave generator {}".format(wave_generator_id), ATTR_WAVETABLE_INTERPOLATION])
        interpolation_type = data_interpolation.get(interpolation_type)
        self.logger.info("Set wave generator interpolation type to {}".format(wave_generator_id))
        output = self.command("WTR {} {} {}\n".format(wave_generator_id, int(wave_table_rate), interpolation_type),
                              returning=False, with_token=True, callback=callback)
        self.get_wave_generator_rate_and_interpolation_type()
        return output

    def get_wave_gen_cycles(self, wave_generator_id, callback=None):
        """
        Gets the number of output cycles set for the given wave generator.
        :param wave_generator_id: int
        :param callback:
        """
        return self.command("WGC? {}\n".format(wave_generator_id + 1), with_token=True, callback=callback)

    def set_wave_gen_cycles(self, wave_generator_id, cycles, callback=None):
        """
        Gets the number of output cycles set for the given wave generator.
        :param wave_generator_id: int
        :param cycles: int
        :param callback:
        """
        self.logger.info("Set wave generator {} cycles to {}".format(wave_generator_id, cycles))
        return self.command("WGC {} {}\n".format(wave_generator_id + 1, cycles), returning=False, with_token=True,
                            callback=callback)

    def get_wave_gen_output_offset(self, wave_generator_id=None, callback=None):
        """
        Reads the current value of the offset which is added to
        the wave generator output (Wave Offset parameter
        value in volatile memory (ID 0x1300010b)). This value
        results either from WOS (p. 97) / SPA (p. 65) / SEP
        (p. 62) settings, or from internal calculation during the
        wave generator output; see WOS for details.
        :param wave_generator_id: int
        :param callback:
        :return:
        """
        if wave_generator_id is None:
            return self.command("WOS?\n", with_token=True, callback=callback)
        else:
            return self.command("WOS? {}\n".format(wave_generator_id + 1), with_token=True, callback=callback)

    def set_wave_gen_output_offset(self, wave_generator_id, offset, callback=None):
        """
        Sets an offset to the output of a wave generator. The
        current wave generator output is then created by
        adding the offset value to the current wave value:
        Generator Output = Offset + Current Wave Value
        :param wave_generator_id:
        :param offset:
        :param callback:
        :return:
        """
        self.logger.info("Set wave generator {} offset to {}".format(wave_generator_id, offset))
        return self.command("WOS {} {}\n".format(wave_generator_id + 1, offset), returning=False, with_token=True,
                            callback=callback)

    def get_connected_table_to_generator(self, wave_generator_id=None, callback=None):
        if wave_generator_id is None:
            return self.command("WSL?\n", with_token=True, callback=callback)
        else:
            return self.command("WSL? {}\n".format(wave_generator_id + 1), with_token=True, callback=callback)

    def set_connected_table_to_generator(self, wave_generator_id, wave_table_id, callback=None):
        self.logger.info("Set connected generator {} to table {}".format(wave_generator_id, wave_table_id))
        return self.command("WSL {} {}\n".format(wave_generator_id + 1, wave_table_id + 1), returning=False,
                            with_token=True,
                            callback=callback)

    def get_configuration_of_trigger_outputs(self, callback=None):
        return self.command("CTO?\n", with_token=True, callback=callback)

    def set_configuration_of_trigger_output(self, trigger_output, parameter, value, callback=None):
        self.logger.info("Set trigger output {} parameter {} = {}".format(trigger_output, parameter, value))
        if parameter == TRIG_OUTPUT_MODE:
            value = trigger_output_mode.get(value)
        parameter = trigger_output_config.get(parameter)
        return self.command("CTO {} {} {}\n".format(trigger_output + 1, parameter, value), returning=False,
                            with_token=True, callback=callback)

    def set_wave_generator_mode(self, generator_id, start_mode, callback=None):
        return self.command("WGO {} {}\n".format(generator_id + 1, start_mode), returning=False, with_token=True,
                            callback=callback)

    def get_wave_generator_mode(self, generator_id=None, callback=None):
        if generator_id is None:
            return self.command("WGO?\n", with_token=True, callback=callback)
        else:
            return self.command("WGO? {}\n".format(generator_id + 1), with_token=True, callback=callback)

    def set_trigger_line(self, trigger_output, trigger_line_params, callback=None):
        """
        :param trigger_output: int
        :param trigger_line_params: list
        :param callback:
        :type trigger_line_params: [(PointNumber, Switch), (...), ...]
        :return:
        """
        sliced = 0
        trigger_appendix = (trigger_output + 1,)
        prepared_command = None
        while sliced <= len(trigger_line_params):
            if prepared_command is not None:
                self.command(prepared_command, returning=False)

            sliced_params = trigger_line_params[sliced: sliced + self.N_TRIG_PER_CMD]
            params = " ".join([" ".join([str(b) for b in trigger_appendix + a]) for a in sliced_params])
            prepared_command = "TWS {}\n".format(params)
            sliced += self.N_TRIG_PER_CMD

        if prepared_command is not None:
            return self.command(prepared_command, returning=False, with_token=True, callback=callback)

    def get_trigger_line(self, trigger_output, start_point=1, number_of_points=None, callback=None):
        if number_of_points is None:
            number_of_points = self.get_value(["Wave generator {}".format(trigger_output), ATTR_WAVETABLE_LENGTH])
        return self.command("TWS? {} {} {}\n".format(start_point, number_of_points, trigger_output + 1),
                            with_token=True, callback=callback)

    def get_wavetable_data(self, wave_table_id, start_point=1, number_of_points=None, callback=None):
        if number_of_points is None:
            number_of_points = self.get_value(["Wave generator {}".format(wave_table_id), ATTR_WAVETABLE_LENGTH])
        return self.command("GWD? {} {} {}\n".format(start_point, number_of_points, wave_table_id + 1), with_token=True,
                            callback=callback)

    def set_wave_form_definition(self, wave_table_id, wave_add_method, wave_type, wave_type_param, callback=None):
        if wave_type == WAVE_TYPE_PNT:
            """
            In this case we need (WaveStartPoint, WaveLength, (WavePoints))
            """
            command_parameter = wave_type_param[0]
            points = wave_type_param[2]

            sliced = 0
            prepared_command = None

            def _make_command(command, command_parameter, sliced_points):
                joined_points = " ".join("%.6f" % point for point in sliced_points)
                return "{} {} {} {}\n".format(command, command_parameter, len(sliced_points), joined_points)

            while sliced <= len(points):
                if prepared_command is not None:
                    self.command(_make_command(prepared_command, command_parameter, sliced_points), returning=False)
                if sliced > 0:
                    wave_add_method = WAVE_TYPE_APPEND

                prepared_command = "WAV %s %s %s" % (wave_table_id + 1, wave_add_method, wave_type)
                sliced_points = points[sliced: sliced + self.N_WAV_POINTS_PER_CMD]
                sliced += self.N_WAV_POINTS_PER_CMD

            if prepared_command is not None:
                return self.command(_make_command(prepared_command, command_parameter, sliced_points), returning=False,
                                    with_token=True, callback=callback)

        elif wave_type == WAVE_TYPE_SINP:
            """
            In this case we need (SegLength, Amp, Offset, WaveLength, StartPoint, CurveCenterPoint)
            """
            command_parameter = "%i %f %f %i %i %i" % wave_type_param
        elif wave_type == WAVE_TYPE_RAMP:
            """
            In this case we need (SegLength, Amp, Offset, WaveLength, StartPoint, SpeedUpDown, CurveCenterPoint)
            """
            command_parameter = "%i %f %f %i %i %i %i" % wave_type_param
        elif wave_type == WAVE_TYPE_LIN:
            """
            In this case we need (SegLength, Amp, Offset, WaveLength, StartPoint, SpeedUpDown)
            """
            command_parameter = "%i %f %f %i %i %i" % wave_type_param

        command = "WAV %s %s %s" % (wave_table_id + 1, wave_add_method, wave_type)
        return self.command("%s %s\n" % (command, command_parameter), returning=False, with_token=True,
                            callback=callback)

    def set_record_table_rate(self, rate, callback=None):
        """
        Sets the record table rate, i.e. the number of servo-loop
        cycles to be used in data recording operations. Settings
        larger than 1 make it possible to cover longer time
        periods.
        :param rate: int
        :param callback:
        :return:
        """
        return self.command("RTR {}\n".format(int(rate)), returning=False, with_token=True, callback=callback)

    def set_recorded_configuration(self, table_id, source, option, callback=None):
        """
        Set data recorder configuration: determines the data
        source and the kind of data (RecordOption) used for
        the given data recorder table.
        :param table_id:
        :param source:
        :param option:
        :param callback:
        :return:
        """
        return self.command("DRC {} {} {}\n".format(table_id + 1, source + 1, option), returning=False, with_token=True,
                            callback=callback)

    def get_recorder_data(self, table_id, start_point=1, number_of_points=None, callback=None):
        if number_of_points is None:
            number_of_points = self.get_value(["Wave generator {}".format(table_id), ATTR_WAVETABLE_LENGTH])
        return self.command("DRR? {} {} {}\n".format(start_point, number_of_points, table_id + 1), with_token=True,
                            callback=callback)

    def set_recorded_trigger_source(self, table_id, source, value, callback=None):
        """
        Set data recorder configuration: determines the data
        source and the kind of data (RecordOption) used for
        the given data recorder table.
        :param table_id:
        :param source:
        :param value:
        :param callback:
        :return:
        """
        return self.command("DRT {} {} {}\n".format(table_id + 1, source, value), returning=False, with_token=True,
                            callback=callback)

    def _move_absolute(self, channel, position, callback=None):
        self.logger.info(u"Move channel {} to absolute position {}".format(channel, position))
        return self.command(u"MOV {} {}\n".format(channel + 1, position), callback, with_token=True,
                            returning=False)

    @expose_method({"generator_id": "0"})
    def start_fly_scan(self, generator_id):
        self.set_wave_generator_mode(int(generator_id), 1)
        self.command("#9\n")

    @expose_method({"channel": "0", "center": "Center of scan", "step_size": "steps size", "pixels": "Scan points",
                    "dwell_time": "usec", "trigger_mode": "0", "interpolating": "0"})
    def set_fly_scan(self, channel=0, center=50, step_size=1, pixels=10, dwell_time=100,
                     trigger_mode=TRIG_MODE_GENERATOR, interpolating=False):
        self.stop_polling()
        self.set_status(STATUS_BUSY)
        channel = int(channel)
        center = float(center)
        step_size = float(step_size)
        pixels = int(pixels)
        dwell_time = float(dwell_time)
        # trigger_mode = int(trigger_mode)
        interpolating = bool(int(interpolating))

        generator_id = None
        for trigger_id in range(self.get_value(ATTR_TRIGGER_OUTPUTS_COUNT)):
            trigger_axis = self.get_value(("Trigger output {}".format(trigger_id), TRIG_OUTPUT_AXIS))
            if trigger_axis == channel:
                generator_id = trigger_id
                break

        if generator_id is None:
            raise DeviceError("Cannot find trigger output bound to selected channel")

        # Clear wave table and all triggers
        self.clear_wave_table(generator_id)
        self.clear_wave_table_triggers()
        # Calculate setpoints, dwell time and offsets
        fly_scan_params = self._prepare_fly_scan_parameters(center, step_size, pixels, dwell_time, trigger_mode)
        dwell_time, start_position, setpoints, pre_scan_pixels, post_scan_pixels, total_pixels, pixels = fly_scan_params
        # Set interpolation and table rate
        interpolation_type = INTERPOLATION_LINE if interpolating else INTERPOLATION_NONE
        wave_table_rate = dwell_time / self.get_value(ATTR_SERVO_UPDATE_TIME)
        self.set_wave_generator_rate_and_interpolation_type(wave_table_rate=wave_table_rate,
                                                            interpolation_type=interpolation_type)
        # Set wave generator cycles
        self.set_wave_gen_cycles(generator_id, 1)

        # Set recorder table rate
        self.set_record_table_rate(int(wave_table_rate))
        self.set_value(["Wave generator {}".format(generator_id), ATTR_WAVETABLE_LENGTH], total_pixels)
        self.set_wave_form_definition(generator_id, WAVE_TYPE_START, WAVE_TYPE_PNT, (1, total_pixels + 50, setpoints))
        # Set generator offset to start position
        self.set_wave_gen_output_offset(generator_id, start_position)
        # Connect generator to wave table
        self.set_connected_table_to_generator(generator_id, generator_id)

        if trigger_mode == TRIG_MODE_GENERATOR:
            self.set_configuration_of_trigger_output(generator_id, TRIG_OUTPUT_MODE, TRIG_MODE_GENERATOR)
            trigger_points_count = int(pixels / 2 + 1)
            trigger_points = []
            for j in range(trigger_points_count):
                trigger_points.append(((pre_scan_pixels + j * 2) + 1, 1))
            self.set_trigger_line(generator_id, trigger_points)
            self.get_trigger_line(generator_id, 1, total_pixels)
        elif trigger_mode == TRIG_MODE_POSITION_DISTANCE:
            self.set_configuration_of_trigger_output(generator_id, TRIG_OUTPUT_MODE, TRIG_MODE_POSITION_DISTANCE)
            self.set_configuration_of_trigger_output(generator_id, TRIG_START, start_position)
            # We need to set stop position to + 1 pixel to stop last trigger
            self.set_configuration_of_trigger_output(generator_id, TRIG_STOP, step_size * (pixels + 1))

        self.set_recorded_configuration(generator_id, generator_id, REC_CURRENT_POSITION)
        self.set_recorded_trigger_source(generator_id, 0, 0)
        self._move_absolute(channel, start_position)

        self.get_wavetable_data(generator_id, 1, total_pixels)
        self.get_wave_gen_output_offset(generator_id)
        self.set_status(STATUS_IDLE)
        self.start_polling()
        self.notify((channel, ATTR_WAVE_GENERATOR), generator_id)
        return generator_id

    def _prepare_fly_scan_parameters(self, center, step_size, pixels, dwell_time, trigger_mode):
        j = 0
        # set motion parameters
        acceleration_speed = 0.02  # the value of acceleration and deceleration
        fly_back_speed = 0.2  # the maximum speed on the way back
        line_dwell = 0  # 12 the time to wait at the beginning of each line (after flyback mainly)
        drift_pixels = 2  # the number of pixels after acceleration and before taking data; 2 pixels should be enough for anything to settle down

        def forHelper(j, k, condition_callback, positions, callback, null_J=True):
            if null_J:
                j = 0
            while condition_callback(j, k):
                positions[k] = callback(j, k)
                j += 1
                k += 1
            return j, k

        if trigger_mode == TRIG_MODE_GENERATOR:  # we need 2 points of the wave generator for a full trigger_mode period, so let's cut every pixel in half
            step_size /= 2
            pixels *= 2
            dwell_time /= 2
            drift_pixels *= 2

        # adjust to max speed
        dwell_time = max(dwell_time, abs(step_size) / self.MAX_SPEED_UMPMS)

        fly_back_speed = min(fly_back_speed, self.MAX_SPEED_UMPMS)
        fly_back_speed = max(fly_back_speed, abs(step_size / dwell_time))

        # The number of pixels we wait at the beginning of each line to let shaking settle down
        LineWaitPixels = int(around(line_dwell / dwell_time))

        # The following ceil will be compensated ("automatically") by a reduced acceleration
        AccelPixels = int(ceil(abs(step_size) / (dwell_time * dwell_time * acceleration_speed)))
        PrescanPixels = LineWaitPixels + AccelPixels + drift_pixels

        FullTravel_um = step_size * (pixels + drift_pixels + AccelPixels)
        Flyback_umpPix = (-1) * sign(step_size) * fly_back_speed * dwell_time
        FlybackAccelPixels = ceil(abs(Flyback_umpPix) / (dwell_time * dwell_time * acceleration_speed))

        if FlybackAccelPixels < abs(FullTravel_um / Flyback_umpPix):
            # The overshot from this ceil will be corrected for in the end
            FlybackPixels = ceil(abs(FullTravel_um / Flyback_umpPix) - FlybackAccelPixels)
            AdjustedFlybackAccelPixels = FlybackAccelPixels
        else:
            FlybackPixels = 0
            # go as close to real value as possible (-> RoundRealToNearestInteger() instead of ceil())
            AdjustedFlybackAccelPixels = around(sqrt(FlybackAccelPixels * abs(FullTravel_um / Flyback_umpPix)))

        PostscanPixels = int(AccelPixels + FlybackPixels + 2 * AdjustedFlybackAccelPixels)
        TotalPixels = int(PrescanPixels + pixels + PostscanPixels)

        Positions_um = [0] * TotalPixels
        fStart_um = center - step_size * (drift_pixels + (pixels + AccelPixels) / 2)

        PostscanPixels = int(AccelPixels + FlybackPixels + 2 * AdjustedFlybackAccelPixels)
        TotalPixels = int(PrescanPixels + pixels + PostscanPixels)

        Positions_um = [0] * TotalPixels
        fStart_um = center - step_size * (drift_pixels + (pixels + AccelPixels) / 2)

        k = 0
        j, k = forHelper(j, k, lambda j, k: j < LineWaitPixels, Positions_um, lambda j, k: 0)
        j, k = forHelper(j, k, lambda j, k: j < AccelPixels, Positions_um,
                         lambda j, k: step_size * (j + 1) * (j + 1) / (2 * AccelPixels))

        TempStart_um = step_size * j * j / (2 * AccelPixels) if AccelPixels != 0 else 0

        j, k = forHelper(j, k, lambda j, k: j < drift_pixels + pixels, Positions_um,
                         lambda j, k: TempStart_um + step_size * (j + 1))
        TempStart_um += step_size * j
        j, k = forHelper(j, k, lambda j, k: j < AccelPixels, Positions_um,
                         lambda j, k: TempStart_um + step_size * (
                                 (j + 1) - (j + 1) * (j + 1) / (2 * AccelPixels)))
        if AccelPixels != 0:
            TempStart_um += step_size * (j - j * j / (2. * AccelPixels))

        # now fly back
        j, k = forHelper(j, k, lambda j, k: j < AdjustedFlybackAccelPixels, Positions_um,
                         lambda j, k: TempStart_um + Flyback_umpPix * (j + 1) * (j + 1) / (
                                 2 * FlybackAccelPixels))

        if FlybackAccelPixels != 0:
            TempStart_um += Flyback_umpPix * j * j / (2 * FlybackAccelPixels)
        j, k = forHelper(j, k, lambda j, k: j < FlybackPixels, Positions_um,
                         lambda j, k: TempStart_um + Flyback_umpPix * (j + 1))
        TempStart_um += Flyback_umpPix * j
        j, k = forHelper(j, k, lambda j, k: j < AdjustedFlybackAccelPixels, Positions_um,
                         lambda j, k: TempStart_um + Flyback_umpPix * (
                                 (j + 1) * 2 * AdjustedFlybackAccelPixels - (j + 1) * (j + 1)) / (
                                              2 * FlybackAccelPixels))

        # Now let's fix possible jumps at the end of unidirectional lines
        # We just find the difference between the start and the last pixel
        Difference = -Positions_um[TotalPixels - 1]
        # and add a fraction of it to all return pixels (cheating!)
        try:
            Difference /= FlybackPixels + 2 * AdjustedFlybackAccelPixels
        except DivisionByZero:
            Difference = 0

        k = PrescanPixels + AccelPixels + pixels
        j = 0
        while j < FlybackPixels + 2 * AdjustedFlybackAccelPixels:
            Positions_um[k + j] += Difference * (j + 1)
            j += 1

        return dwell_time, fStart_um, Positions_um, PrescanPixels, PostscanPixels, TotalPixels, pixels


class DevicePiPort(DevicePort, DevicePi):
    response_timeout = DevicePi.response_timeout
    terminator = DevicePi.terminator
    push_commands_max = DevicePi.push_commands_max

    def __init__(self, port, baud_rate=115200, parity=serial.PARITY_NONE, stop_bits=serial.STOPBITS_ONE,
                 byte_size=serial.EIGHTBITS, device_id=None, config=None):
        self.channel_mapping = bidict.MutableBidict()
        self.valid_commands = []
        super(DevicePiPort, self).__init__(port, baud_rate, parity, stop_bits, byte_size, device_id, config)

    def handshake(self):
        self.logger.info(u"Handshake initiated")
        self.push(b"ERR?\n")
        time.sleep(0.05)  # Give some time for devices to reply
        if self.serial_port.in_waiting != 0:
            output = self.read_all().strip()
            is_error_message = self.error_pattern.match(output)
            if is_error_message is not None:
                return True
        return False


class DevicePiChannel(DeviceChannel):

    def __init__(self, device, channel, device_id=None, config=None):
        self.channel_id = None
        super().__init__(device, channel, device_id, config)

    def _init_attributes(self):
        super(DevicePiChannel, self)._init_attributes()
        self.create_attribute(ATTR_ENCODER_BASE_UNIT, readonly=True)
        self.create_attribute(ATTR_POSITION, default_type=np.float64, decimals=9,
                                     set_function=self._move_absolute, set_value_when_set_function=False)
        self.create_attribute(ATTR_POSITION_UPPER_LIMIT, default_type=np.float64, decimals=9,
                                                 set_function=self.set_position_upper_limit)
        self.create_attribute(ATTR_POSITION_LOWER_LIMIT, default_type=np.float64, decimals=9,
                                                 set_function=self.set_position_lower_limit)
        self.create_attribute(ATTR_VELOCITY, default_type=np.float64, decimals=9, set_function=self.set_velocity)
        self.create_attribute(ATTR_SERVO_STATE, default_type=np.bool, set_function=self.set_servo_state)
        self.create_attribute(ATTR_REFERENCED, readonly=True, default_type=np.bool)
        self.create_attribute(ATTR_MOVING, readonly=True, default_type=np.bool)
        self.create_attribute(ATTR_SCANNING, readonly=True, default_type=np.bool)
        self.create_attribute(ATTR_WAVE_GENERATOR, readonly=True, default_type=np.uint16)
        self.create_attribute(ATTR_RECORD_TABLE_DATA, readonly=True)

    def handle_configuration(self):
        if self.configured:
            return
        start_at = time.time()

        def _finish_configuration(*_, **__):
            self._config_commands()
            self._config_attributes()
            self.configured = True
            self.start_polling()
            self.set_status(STATUS_CONFIGURED)
            self.logger.info(u"Device configuration took {} sec.".format(time.time() - start_at))

        self.connected = True
        self.channel_id = self.device.channel_mapping.inv.get(self.channel)
        self.command("PUN? {}\n".format(self.channel_id))
        if "FRF?" in self.device.valid_commands:
            self.command("FRF? {}\n".format(self.channel_id))
        else:
            self.set_value(ATTR_REFERENCED, True)
        self.command("SVO? {}\n".format(self.channel_id))
        self.get_position_lower_limit()
        self.get_position_upper_limit()
        self.get_velocity(callback=_finish_configuration)

    def subject_update(self, key, value, subject):
        super(DevicePiChannel, self).subject_update(key, value, subject)

        if self.connected and isinstance(key, tuple) and key[0] == self.channel:
            attribute = key[1]

            if attribute == ATTR_ENCODER_BASE_UNIT:
                self.set_attribute((ATTR_POSITION, UNIT), value)
                self.set_attribute((ATTR_POSITION_UPPER_LIMIT, UNIT), value)
                self.set_attribute((ATTR_POSITION_LOWER_LIMIT, UNIT), value)
                self.set_attribute((ATTR_RECORD_TABLE_DATA, UNIT), value)
                self.set_attribute((ATTR_VELOCITY, UNIT), "{}/s".format(value))
            elif attribute == ATTR_POSITION_UPPER_LIMIT:
                self.set_attribute((ATTR_POSITION, MAX), value)
            elif attribute == ATTR_POSITION_LOWER_LIMIT:
                self.set_attribute((ATTR_POSITION, MIN), value)
            elif attribute == ATTR_MOVING:
                if self.get_value(ATTR_SCANNING) or value:
                    self.set_status(STATUS_BUSY)
                else:
                    self.set_status(STATUS_IDLE)
            elif attribute == ATTR_SCANNING:
                if self.get_value(ATTR_MOVING) or value:
                    self.set_status(STATUS_BUSY)
                else:
                    self.set_status(STATUS_IDLE)

            self.set_raw_value(ATTR_LATENCY, self.device.get_value(ATTR_LATENCY))
            self.set_raw_value(attribute, value)

    @expose_method({"Field_size": ATTR_POSITION, "Steps": "Number of steps", "Dwell_time": "milliseconds",
                    "Interpolating": "bool"})
    def set_relative_fly_scan(self, Field_size, Steps, Dwell_time, Interpolating):
        filed_size = device_units(self, ATTR_POSITION, Field_size).m
        step_size = filed_size / int(Steps)
        actual_position = self.get_value(ATTR_POSITION)
        generator_id = self.device.set_fly_scan(self.channel, actual_position, step_size, int(Steps), float(Dwell_time),
                                                interpolating=int(Interpolating))
        self.set_value(ATTR_WAVE_GENERATOR, generator_id)
        return generator_id

    @expose_method()
    def start_fly_scan(self):
        generator_id = self.get_value(ATTR_WAVE_GENERATOR)
        assert isinstance(generator_id, int)
        self.device.start_fly_scan(generator_id)

    def get_position_upper_limit(self, callback=None):
        return self.command(u"TMN? {}\n".format(self.channel_id), callback=callback, with_token=True)

    def get_position_lower_limit(self, callback=None):
        return self.command(u"TMX? {}\n".format(self.channel_id), callback=callback, with_token=True)

    def set_position_upper_limit(self, position, callback=None):
        self.logger.info(u"Set upper position limit to {}".format(position))
        return self.set_attribute((ATTR_POSITION, MAX), position, callback=callback)

    def set_position_lower_limit(self, position, callback=None):
        self.logger.info(u"Set lower position limit to {}".format(position))
        return self.set_attribute((ATTR_POSITION, MIN), position, callback=callback)

    def get_velocity(self, callback=None):
        return self.command(u"VEL? {}\n".format(self.channel_id), callback=callback, with_token=True)

    def set_velocity(self, velocity, callback=None):
        self.logger.info(u"Set velocity to {}".format(velocity))
        return self.command(
            u"VEL {} {}\n".format(self.channel_id, velocity), callback=callback, with_token=True, returning=False)

    def _move_absolute(self, position, callback=None):
        self.logger.info(u"Move to absolute position {}".format(position))
        return self.command(u"MOV {} {}\n".format(self.channel_id, position), callback, with_token=True,
                            returning=False)

    @expose_method()
    def stop(self, callback=None):
        self.logger.info(u"Stop movement")
        return self.command(u"STP\n", callback, with_token=True, returning=False)

    @expose_method()
    def find_reference(self, callback=None):
        self.logger.info(u"Reference")
        return self.command(u"FRF{}\n".format(self.channel_id), callback, with_token=True, returning=False)

    @expose_method({"position": ATTR_POSITION})
    def move_absolute(self, position):
        try:
            position = device_units(self, ATTR_POSITION, position)
            self.logger.info(u"Move motor channel to absolute position {}".format(position))
            return self.set_value(ATTR_POSITION, position.m)
        except (UndefinedUnitError, ValueError) as e:
            self.logger.exception(e)
            raise DeviceUnitError(e)

    @expose_method({"position": ATTR_POSITION})
    def move_relative(self, position):
        try:
            position = device_units(self, ATTR_POSITION, position)
            self.logger.info(u"Move by relative position {}".format(position))
            return self.set_value(ATTR_POSITION, self[ATTR_POSITION].value() + position)
        except (UndefinedUnitError, ValueError) as e:
            self.logger.exception(e)
            raise DeviceUnitError(e)

    def set_servo_state(self, flag, callback=None):
        self.logger.info(u"Set servo state to {}".format(flag))
        return self.command(u"SVO {} {}\n".format(self.channel_id, int(flag)), callback, with_token=True,
                            returning=False)
